"""Utility functions for creating regularizer object."""

AVAILABLE_REGULARIZERS = ["UnRegularized", "Ridge", "Lasso", "GroupLasso"]


def create_regularizer(name: str):
    """
    Create a regularizer from a given name.

    Parameters
    ----------
    name :
        The string name of the regularizer to create. Must be one of: 'UnRegularized', 'Ridge', 'Lasso', 'GroupLasso'.

    Returns
    -------
    :
        The regularizer instance with default parameters.

    Raises
    ------
    ValueError
        If the `name` provided does not match to any available regularizer.
    """
    if name == "UnRegularized":
        from .regularizer import UnRegularized

        return UnRegularized()
    elif name == "Ridge":
        from .regularizer import Ridge

        return Ridge()
    elif name == "Lasso":
        from .regularizer import Lasso

        return Lasso()
    elif name == "GroupLasso":
        from .regularizer import GroupLasso

        return GroupLasso()
    else:
        raise ValueError(
            f"Unknown regularizer: {name}. "
            f"Regularizer must be one of {AVAILABLE_REGULARIZERS}"
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import nemos
import sys, os
from pathlib import Path

sys.path.insert(0, str(Path('..', 'src').resolve()))
sys.path.insert(0, os.path.abspath('sphinxext'))


project = 'nemos'
copyright = '2024'
author = 'E Balzani'
version = release = nemos.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# The Root document
root_doc = "index"

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',  # Links to source code
    'sphinx.ext.doctest',
    'sphinx_copybutton',  # Adds copy button to code blocks
    'sphinx_design',  # For layout components
    'myst_nb',
    'sphinx_contributors',
    'sphinx_code_tabs',
    'sphinx.ext.mathjax',
    'sphinx_autodoc_typehints',
    'sphinx_togglebutton',
    'matplotlib.sphinxext.plot_directive',
    "matplotlib.sphinxext.mathmpl",
]

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "dollarmath",
    "html_admonition",
    "html_image",
]

templates_path = ['_templates']
exclude_patterns = ['_build', "docstrings", 'Thumbs.db', 'nextgen', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


# Generate the API documentation when building
autosummary_generate = True
numpydoc_show_class_members = True
autodoc_default_options = {
    'members': True,
    'inherited-members': True,
    'undoc-members': True,
    'show-inheritance': True,
    'special-members': ' __add__, __mul__, __pow__'
}

# # napolean configs
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

autodoc_typehints = "description"  # Use "description" to place hints in the description
autodoc_type_aliases = {
    "ArrayLike": "ArrayLike",
    "NDArray": "NDArray",
    "TsdFrame": "pynapple.TsdFrame",
    "JaxArray": "JaxArray",
}
autodoc_typehints_format = "short"

numfig = True

html_theme = 'pydata_sphinx_theme'

html_favicon = "assets/NeMoS_favicon.ico"

# Additional theme options
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/flatironinstitute/nemos/",
            "icon": "fab fa-github",
            "type": "fontawesome",
        },
        {
            "name": "X",
            "url": "https://x.com/nemos_neuro",
            "icon": "fab fa-square-x-twitter",
            "type": "fontawesome",
        },
    ],
    "show_prev_next": True,
    "header_links_before_dropdown": 6,
    "navigation_depth": 3,
    "logo": {
      "image_light": "_static/NeMoS_Logo_CMYK_Full.svg",
      "image_dark": "_static/NeMoS_Logo_CMYK_White.svg",
   },
    "secondary_sidebar_items": {
        "**": ["page-toc", "sourcelink"],
        "background/basis/README": [],
    },
}

html_sidebars = {
    "index": [],
    "installation":[],
    "quickstart": [],
    "background/README": [],
    "how_to_guide/README": [],
    "tutorials/README": [],
    "**": ["search-field.html", "sidebar-nav-bs.html"],
}


# Path for static files (custom stylesheets or JavaScript)
html_static_path = ['assets/stylesheets', "assets"]
html_css_files = ['custom.css']

html_js_files = [
    "https://code.iconify.design/2/2.2.1/iconify.min.js"
]

# Copybutton settings (to hide prompt)
copybutton_prompt_text = r">>> |\$ |# "
copybutton_prompt_is_regexp = True

sphinxemoji_style = 'twemoji'

nb_execution_timeout = 60 * 15  # Set timeout in seconds (e.g., 15 minutes)

nitpicky = True

# Get exclusion patterns from an environment variable
exclude_tutorials = os.environ.get("EXCLUDE_TUTORIALS", "false").lower() == "true"

if exclude_tutorials:
    nb_execution_excludepatterns = ["tutorials/**", "how_to_guide/**", "background/**"]

viewcode_follow_imported_members = True

# option for mpl extension
plot_html_show_formats = False

# raise an error if exec error in notebooks
nb_execution_raise_on_error = True


"""Utilities for manipulating and checking PyTrees."""

import operator
from functools import partial, reduce
from typing import Any, Callable, Optional

import jax
import jax.numpy as jnp


def _get_not_inf(array: jnp.ndarray) -> jnp.ndarray:
    """
    Identify non-infinite entries within an array.

    This function evaluates each element in the input array to determine whether it is finite (not infinite).
    It performs this check across all axes except the first one, aggregating results using a logical 'AND' operation.
    Thus, for a given element along the first axis, if all corresponding elements in other dimensions are finite,
    the result is True; otherwise, it is False.

    Parameters
    ----------
    array : jnp.ndarray
        Input array to check for infinite values.

    Returns
    -------
        A 1D boolean array of length equal to the size of the first dimension of the input array.
        Each entry in this array corresponds to an aggregation across all other dimensions of the input array,
        with True indicating all values are finite (not infinite) and False indicating at least one infinite value.
    """
    return jax.numpy.all(~jnp.isinf(array), axis=range(1, array.ndim))


def _get_not_nan(array: jnp.ndarray) -> jnp.ndarray:
    """
    Identify non-NaN (Not a Number) entries within an array.

    Similar to the _get_not_inf function, this function checks each element in the input array for being non-NaN across
    all axes except the first one. It aggregates these checks using a logical 'AND' operation. An element along the
    first axis is considered valid (True) if all its corresponding elements in other dimensions are not NaN.

    Parameters
    ----------
    array :
        The input array for which to check for NaN values.

    Returns
    -------
    :
        A 1D boolean array of length equal to the size of the first dimension of the input array.
        Each entry in this array is the result of an aggregation across all other dimensions,
        with True indicating all values are not NaN, and False indicating at least one NaN value.
    """
    return jax.numpy.all(~jnp.isnan(array), axis=range(1, array.ndim))


def _get_valid_tree(tree: Any) -> jnp.ndarray:
    """
    Filter valid entries across all leaves in a pytree.

    Processes a pytree to identify entries without NaN or infinite values across all its leaves. A leaf is considered
    valid if all its entries are finite and not NaN. The function assumes homogeneous first dimension across all leaves.

    Parameters
    ----------
    tree :
        A pytree with leaves as NDArrays sharing the same size for the first dimension.

    Returns
    -------
        A boolean array indicating validity of each entry across all leaves. True represents a valid entry,
        while False indicates an invalid (NaN or infinite) entry.
    """
    valid = jax.tree_util.tree_leaves(
        jax.tree_util.tree_map(lambda x: _get_not_inf(x) & _get_not_nan(x), tree)
    )
    return reduce(jnp.logical_and, valid)


def get_valid_multitree(*tree: Any) -> jnp.ndarray:
    """
    Filter valid entries across multiple pytrees.

    Evaluates multiple pytrees to identify common entries that are valid (non-NaN and finite) across all of them.
    Assumes that all pytrees have leaves with NDArrays sharing the same size for the first dimension.

    Parameters
    ----------
    tree :
        Variable number of pytrees with NDArrays as leaves, each having a consistent first dimension size.

    Returns
    -------
    :
        A boolean array indicating the validity of each entry across all leaves in all pytrees. True for valid entries,
        False for invalid ones.
    """
    return reduce(jnp.logical_and, map(_get_valid_tree, tree))


def pytree_map_and_reduce(
    map_fn: Callable,
    reduce_fn: Callable,
    *pytrees: Any,
    is_leaf: Optional[Callable[[Any], bool]] = None,
):
    """
    Apply a mapping function to each leaf of the pytrees and then reduce the results.

    This function performs a map/reduce operation where a mapping function is applied
    to each leaf of the given pytrees, and then a reduction function is used to
    aggregate these results into a single output.

    Parameters
    ----------
    map_fn :
        A function to be applied to each leaf of the pytrees. This function should
        take a single argument and return a single value.
    reduce_fn :
        A function that reduces the mapped results. This function should take an
        iterable and return a single value.
    *pytrees :
        One or more pytrees to which the map and reduce functions are applied.
    is_leaf :
        Callable, returns true if sub-tree is a leaf.

    Returns
    -------
    The result of applying the reduce function to the mapped results. The type of the
    return value depends on the reduce function.

    Examples
    --------
    >>> import nemos as nmo
    >>> pytree1 = nmo.pytrees.FeaturePytree(a=jnp.array([0]), b=jnp.array([0]))
    >>> pytree2 = nmo.pytrees.FeaturePytree(a=jnp.array([10]), b=jnp.array([20]))
    >>> map_fn = lambda x, y: x > y
    >>> # Example usage
    >>> result_any = pytree_map_and_reduce(map_fn, any, pytree1, pytree2)
    """
    cond_tree = jax.tree_util.tree_map(map_fn, *pytrees, is_leaf=is_leaf)
    # for some reason, tree_reduce doesn't work well with any.
    return reduce_fn(jax.tree_util.tree_leaves(cond_tree))


def tree_slice(data: Any, idx, is_leaf: Optional[Callable] = None):
    """
    Apply an indexing operation to each array in a nested structure.

    Parameters
    ----------
    data :
        A nested structure containing arrays (e.g., a dictionary of arrays).
    idx :
        The indexing operation to apply. This can be an integer, slice,
        NumPy array (boolean or integer), tuple of indexing operations, ellipsis, or None.
    is_leaf :
        Callable, returns true if sub-tree is a leaf.

    Returns
    -------
    Any
        A nested structure with the same format as `data`, where each array has been sliced according to `idx`.
    """
    return jax.tree_util.tree_map(lambda x: x[idx], data, is_leaf=is_leaf)


# The following functions are adapted from jaxopt.tree_utils

tree_add = partial(jax.tree_util.tree_map, operator.add)
tree_add.__doc__ = "Tree addition."

tree_sub = partial(jax.tree_util.tree_map, operator.sub)
tree_sub.__doc__ = "Tree subtraction."


def tree_scalar_mul(scalar, tree_x):
    """Compute scalar * tree_x."""
    return jax.tree_util.tree_map(lambda x: scalar * x, tree_x)


def tree_add_scalar_mul(tree_x, scalar, tree_y):
    """Compute tree_x + scalar * tree_y."""
    return jax.tree_util.tree_map(lambda x, y: x + scalar * y, tree_x, tree_y)


def tree_sum(tree_x):
    """Compute sum(tree_x)."""
    sums = jax.tree_util.tree_map(jnp.sum, tree_x)
    return jax.tree_util.tree_reduce(operator.add, sums)


def tree_l2_norm(tree_x, squared=False):
    """Compute the l2 norm ||tree_x||."""
    squared_tree = jax.tree_util.tree_map(
        lambda leaf: jnp.square(leaf.real) + jnp.square(leaf.imag), tree_x
    )
    sqnorm = tree_sum(squared_tree)
    if squared:
        return sqnorm
    else:
        return jnp.sqrt(sqnorm)


def tree_zeros_like(tree_x):
    """Creates an all-zero tree with the same structure as tree_x."""
    return jax.tree_util.tree_map(jnp.zeros_like, tree_x)



import abc
import inspect
from typing import List, Tuple


def reimplements_method(
    class_obj: type, base_class_obj: type, method_name: str
) -> bool:
    """
    Check if a class has a specific method defined in the subclass, not inherited from the specified base class.

    Parameters
    ----------
    class_obj :
        The class object to check.
    base_class_obj :
        The base class object against which the check is performed.
    method_name : str
        The name of the method to check for.

    Returns
    -------
    bool
        True if the method is defined in the subclass and not inherited from the base class, False otherwise.
    """

    subclass_method = getattr(class_obj, method_name, None)
    superclass_method = getattr(base_class_obj, method_name, None)
    return subclass_method != superclass_method


def get_subclass_methods(class_obj: type) -> List[Tuple[str, type]]:
    """
    Get the subclass-only methods.

    Parameters
    ----------
    class_obj :
        The class object to inspect.

    Returns
    -------
    :
        A list of tuples representing the methods that are specific to the subclass.
        Each tuple contains the method name (str) and the corresponding method object.
    """
    class_methods = inspect.getmembers(class_obj, predicate=inspect.isfunction)

    # Retrieve the superclass methods
    superclass_methods = []
    if class_obj.__bases__:
        superclass = class_obj.__bases__[0]
        superclass_methods = inspect.getmembers(
            superclass, predicate=inspect.isfunction
        )

    # Filter out the methods defined in the superclass
    subclass_methods = [
        method for method in class_methods if method not in superclass_methods
    ]

    return subclass_methods


def list_abstract_methods(class_obj: type) -> List[Tuple[str, type]]:
    """
    Retrieve the names of abstract methods from a class object.

    Parameters
    ----------
    class_obj :
        The class object to inspect.

    Returns
    -------
        A list of tuples representing the abstract methods in the class.
        Each tuple contains the method name (str) and the corresponding method object.
    """
    class_methods = get_subclass_methods(class_obj)
    abstract_methods = [
        (method_name, method)
        for method_name, method in class_methods
        if getattr(method, "__isabstractmethod__", False)
    ]
    return abstract_methods


def is_abstract(class_obj: type) -> bool:
    """
    Check if a type object is an abstract class.

    Parameters
    ----------
    class_obj : type
        The type object to check.

    Returns
    -------
    bool
        True if the type object is an abstract class, False otherwise.
    """
    return abc.ABC in getattr(class_obj, "__bases__", [])


def get_non_abstract_classes(module) -> List[Tuple[str, type]]:
    """
    List all non-abstract classes in a module.

    Parameters
    ----------
    module : module
        The module object to inspect.

    Returns
    -------
    :
        A list of tuples representing the non-abstract classes in the module.
        Each tuple contains the class name (str) and the corresponding class object.
    """
    basis_classes = inspect.getmembers(
        module, lambda obj: inspect.isclass(obj) and obj.__module__ == module.__name__
    )
    return [
        (name, class_obj)
        for name, class_obj in basis_classes
        if not is_abstract(class_obj)
    ]


def get_abstract_classes(module) -> List[Tuple[str, type]]:
    """
    List all abstract classes in a module.

    Parameters
    ----------
    module : module
        The module object to inspect.

    Returns
    -------
    List[Tuple[str, type]]
        A list of tuples representing the abstract classes in the module.
        Each tuple contains the class name (str) and the corresponding class object.
    """
    basis_classes = inspect.getmembers(
        module, lambda obj: inspect.isclass(obj) and obj.__module__ == module.__name__
    )
    return [
        (name, class_obj) for name, class_obj in basis_classes if is_abstract(class_obj)
    ]


def get_superclass_abstract_methods(class_obj: type) -> List[Tuple[str, type, type]]:
    """
    Retrieve the abstract methods defined in the superclass.

    Parameters
    ----------
    class_obj :
        The class object to inspect.

    Returns
    -------
    List[Tuple[str, type, type]]
        A list of tuples representing the abstract methods defined in the superclass.
        Each tuple contains the method name (str), method object, and superclass object.
    """
    super_class = class_obj.__base__

    if super_class:
        super_class_abstract_methods = get_superclass_abstract_methods(super_class)
        super_class_abstract_methods += [
            (method_name, method, super_class)
            for method_name, method in list_abstract_methods(super_class)
            if getattr(method, "__isabstractmethod__", False)
        ]
    else:
        super_class_abstract_methods = []

    return super_class_abstract_methods


def check_all_abstract_methods_compliance(module) -> None:
    """
    Check if all classes in a module properly implement abstract methods defined in their respective base classes.

    Parameters
    ----------
    module :
        The module object to inspect.

    Raises
    ------
    ValueError
        If any abstract method is not implemented in the subclass.
    """
    non_abstract_classes = get_non_abstract_classes(module)
    for _, base_class in non_abstract_classes:
        string = f"\nAnalyzing class {base_class.__name__}:\n"
        string += "-" * len(string)
        print(string)
        for method_name, method, super_class in get_superclass_abstract_methods(
            base_class
        ):
            boolean = reimplements_method(base_class, super_class, method_name)
            print(f'Is method "{method_name}" re-instantiated? {boolean}')
            if not boolean:
                raise ValueError(
                    f"Abstract method {method} not implemented in {base_class} sub-class!"
                )


def trim_kwargs(cls: type, kwargs: dict, class_specific_params: dict):
    """
    Filter a dictionary of keyword arguments to include only those specific to a given class.

    Parameters
    ----------
    cls :
        The class object for which the keyword arguments are filtered.
    kwargs :
        A dictionary of keyword arguments to be filtered.
    class_specific_params :
        A mapping of class names to sets or lists of allowed parameter names.

    Returns
    -------
    :
        A dictionary containing only the keyword arguments specific to the given class.

    Example
    -------
    >>> class_specific_params = {
    ...     'MyClass': {'param1', 'param2'},
    ...     'OtherClass': {'param3', 'param4'}
    ... }
    >>> kwargs = {'param1': 10, 'param3': 20, 'param5': 30}
    >>> class MyClass:
    ...     pass
    >>> trim_kwargs(MyClass, kwargs, class_specific_params)
    {'param1': 10}
    """
    return {
        key: value
        for key, value in kwargs.items()
        if key in class_specific_params[cls.__name__]
    }



"""
Testing configurations for the NeMoS library.

This module contains test fixtures required to set up and verify the functionality
of the modules of the NeMoS library.

Note:
    This module primarily serves as a utility for test configurations, setting up initial conditions,
    and loading predefined parameters for testing various functionalities of the NeMoS library.
"""

import abc

import jax
import jax.numpy as jnp
import numpy as np
import pynapple as nap
import pytest

import nemos as nmo
import nemos._inspect_utils as inspect_utils
import nemos.basis.basis as basis
from nemos.basis import AdditiveBasis, MultiplicativeBasis
from nemos.basis._basis import Basis

# shut-off conversion warnings
nap.nap_config.suppress_conversion_warnings = True


@pytest.fixture()
def basis_class_specific_params():
    """Returns all the params for each class."""
    all_cls = list_all_basis_classes("Conv") + list_all_basis_classes("Eval")
    return {cls.__name__: cls._get_param_names() for cls in all_cls}


class BasisFuncsTesting(abc.ABC):
    """
    An abstract base class that sets the foundation for individual basis function testing.
    This class requires an implementation of a 'cls' method, which is utilized by the meta-test
    that verifies if all basis functions are properly tested.
    """

    @abc.abstractmethod
    def cls(self):
        pass


class CombinedBasis(BasisFuncsTesting):
    """
    This class is used to run tests on combination operations (e.g., addition, multiplication) among Basis functions.

    Properties:
    - cls: Class (default = None)
    """

    cls = None

    @staticmethod
    def instantiate_basis(
        n_basis, basis_class, class_specific_params, window_size=10, **kwargs
    ):
        """Instantiate and return two basis of the type specified."""

        # Set non-optional args
        default_kwargs = {
            "n_basis_funcs": n_basis,
            "window_size": window_size,
            "decay_rates": np.arange(1, 1 + n_basis),
        }
        repeated_keys = set(default_kwargs.keys()).intersection(kwargs.keys())
        if repeated_keys:
            raise ValueError(
                "Cannot set `n_basis_funcs, window_size, decay_rates` with kwargs"
            )

        # Merge with provided  extra kwargs
        kwargs = {**default_kwargs, **kwargs}

        if basis_class == AdditiveBasis:
            kwargs_mspline = inspect_utils.trim_kwargs(
                basis.MSplineEval, kwargs, class_specific_params
            )
            kwargs_raised_cosine = inspect_utils.trim_kwargs(
                basis.RaisedCosineLinearConv, kwargs, class_specific_params
            )
            b1 = basis.MSplineEval(**kwargs_mspline)
            b2 = basis.RaisedCosineLinearConv(**kwargs_raised_cosine)
            basis_obj = b1 + b2
        elif basis_class == MultiplicativeBasis:
            kwargs_mspline = inspect_utils.trim_kwargs(
                basis.MSplineEval, kwargs, class_specific_params
            )
            kwargs_raised_cosine = inspect_utils.trim_kwargs(
                basis.RaisedCosineLinearConv, kwargs, class_specific_params
            )
            b1 = basis.MSplineEval(**kwargs_mspline)
            b2 = basis.RaisedCosineLinearConv(**kwargs_raised_cosine)
            basis_obj = b1 * b2
        else:
            basis_obj = basis_class(
                **inspect_utils.trim_kwargs(basis_class, kwargs, class_specific_params)
            )
        return basis_obj


# automatic define user accessible basis and check the methods
def list_all_basis_classes(filter_basis="all") -> list[type]:
    """
    Return all the classes in nemos.basis which are a subclass of Basis,
    which should be all concrete classes except TransformerBasis.
    """
    all_basis = [
        class_obj
        for _, class_obj in inspect_utils.get_non_abstract_classes(basis)
        if issubclass(class_obj, Basis)
    ] + [
        bas
        for _, bas in inspect_utils.get_non_abstract_classes(nmo.basis._basis)
        if bas != basis.TransformerBasis
    ]
    if filter_basis != "all":
        all_basis = [a for a in all_basis if filter_basis in a.__name__]
    return all_basis


# Sample subclass to test instantiation and methods
class MockRegressor(nmo.base_regressor.BaseRegressor):
    """
    Mock implementation of the BaseRegressor abstract class for testing purposes.
    Implements all required abstract methods as empty methods.
    """

    def __init__(self, std_param: int = 0):
        """Initialize a MockBaseRegressor instance with optional standard parameters."""
        self.std_param = std_param
        super().__init__()

    def fit(self, X, y):
        pass

    def predict(self, X) -> jnp.ndarray:
        pass

    def score(
        self,
        X,
        y,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def simulate(
        self,
        random_key: jax.Array,
        feed_forward_input,
        **kwargs,
    ):
        pass

    def _check_params(self, *args, **kwargs):
        pass

    def _check_input_and_params_consistency(self, *args, **kwargs):
        pass

    def _check_input_dimensionality(self, *args, **kwargs):
        pass

    def _get_coef_and_intercept(self):
        pass

    def _set_coef_and_intercept(self, params):
        pass

    def update(self, *args, **kwargs):
        pass

    def initialize_state(self, *args, **kwargs):
        pass

    def initialize_params(self, *args, **kwargs):
        pass

    def _predict_and_compute_loss(self, params, X, y):
        pass

    def _get_optimal_solver_params_config(self):
        return None, None, None


class MockRegressorNested(MockRegressor):
    def __init__(self, other_param: int, std_param: int = 0):
        super().__init__(std_param=std_param)
        self.other_param = MockGLM(std_param=other_param)


class MockGLM(nmo.glm.GLM):
    """
    Mock implementation of the BaseRegressor abstract class for testing purposes.
    Implements all required abstract methods as empty methods.
    """

    def __init__(self, std_param: int = 0):
        """Initialize a MockBaseRegressor instance with optional standard parameters."""
        self.std_param = std_param
        super().__init__()

    def fit(self, X, y):
        pass

    def predict(self, X) -> jnp.ndarray:
        pass

    def score(
        self,
        X,
        y,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def simulate(
        self,
        random_key: jax.Array,
        feed_forward_input,
        **kwargs,
    ):
        pass

    def _get_coef_and_intercept(self):
        pass

    def _set_coef_and_intercept(self, params):
        pass


@pytest.fixture
def mock_regressor():
    return MockRegressor(std_param=2)


@pytest.fixture
def mock_regressor_nested():
    return MockRegressorNested(other_param=1, std_param=2)


@pytest.fixture
def mock_glm():
    return MockGLM(std_param=2)


@pytest.fixture
def poissonGLM_model_instantiation():
    """Set up a Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    observation_model = nmo.observation_models.PoissonObservations(jnp.exp)
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate


@pytest.fixture
def poisson_population_GLM_model():
    """Set up a population Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(500, 5))
    b_true = -2 * np.ones((3,))
    w_true = np.random.normal(size=(5, 3))
    observation_model = nmo.observation_models.PoissonObservations(jnp.exp)
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model, regularizer=regularizer
    )
    rate = jnp.exp(jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate


@pytest.fixture
def poisson_population_GLM_model_pytree(poisson_population_GLM_model):
    """Set up a population Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = poisson_population_GLM_model
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model, regularizer=model.regularizer
    )
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


@pytest.fixture
def poissonGLM_model_instantiation_pytree(poissonGLM_model_instantiation):
    """Set up a Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = poissonGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(model.observation_model, model.regularizer)
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


@pytest.fixture
def coupled_model_simulate():
    """Set up a Poisson GLM from a predefined configuration in a json file.

    This fixture reads parameters for a Poisson GLM from a json configuration file, initializes
    the model accordingly, and returns the model instance with other related parameters.

    Returns:
        tuple: A tuple containing:
            - the coupling coeffs
            - the feedforward coeffs
            - the intercepts
            - jax.random.key(123) (jax.Array): A pseudo-random number generator key.
            - feedforward_input (jax.numpy.ndarray): Feedforward input values from the config.
            - coupling_basis (jax.numpy.ndarray): Coupling basis values from the config.
            - init_spikes (jax.numpy.ndarray): Initial spike values from the config.
            - a link function.

    """

    n_neurons, coupling_duration, sim_duration = 2, 100, 1000
    coupling_filter_bank = np.zeros((coupling_duration, n_neurons, n_neurons))
    for unit_i in range(n_neurons):
        for unit_j in range(n_neurons):
            coupling_filter_bank[:, unit_i, unit_j] = (
                nmo.simulation.difference_of_gammas(coupling_duration)
            )
    # shrink the filters for simulation stability
    coupling_filter_bank *= 0.8
    basis = nmo.basis.RaisedCosineLogEval(20)

    # approximate the coupling filters in terms of the basis function
    _, coupling_basis = basis.evaluate_on_grid(coupling_filter_bank.shape[0])
    coupling_coeff = nmo.simulation.regress_filter(coupling_filter_bank, coupling_basis)
    feedforward_coeff = np.ones((n_neurons, 2))
    intercepts = -3 * jnp.ones(n_neurons)

    feedforward_input = jnp.c_[
        jnp.cos(jnp.linspace(0, np.pi * 4, sim_duration)),
        jnp.sin(jnp.linspace(0, np.pi * 4, sim_duration)),
    ]

    feedforward_input = jnp.tile(feedforward_input[:, None], (1, n_neurons, 1))
    init_spikes = jnp.zeros((coupling_duration, n_neurons))

    return (
        coupling_coeff,
        feedforward_coeff,
        intercepts,
        jax.random.key(123),
        feedforward_input,
        coupling_basis,
        init_spikes,
        jnp.exp,
    )


@pytest.fixture
def jaxopt_solvers():
    return [
        "GradientDescent",
        "BFGS",
        "LBFGS",
        "ScipyMinimize",
        "NonlinearCG",
        "ScipyBoundedMinimize",
        "LBFGSB",
        "ProximalGradient",
    ]


@pytest.fixture
def group_sparse_poisson_glm_model_instantiation():
    """Set up a Poisson GLM for testing purposes with group sparse weights.

    This fixture initializes a Poisson GLM with random, group sparse, parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    w_true[1:4] = 0.0
    mask = np.zeros((2, 5))
    mask[0, 1:4] = 1
    mask[1, [0, 4]] = 1
    observation_model = nmo.observation_models.PoissonObservations(jnp.exp)
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate, mask


@pytest.fixture
def group_sparse_poisson_glm_population():
    """Set up a Poisson GLM for testing purposes with group sparse weights.

    This fixture initializes a Poisson GLM with random, group sparse, parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((3,))
    w_true = np.random.normal(size=(5, 3))
    w_true[1:4, 0] = 0.0
    mask = np.zeros((2, 5))
    mask[0, 1:4] = 1
    mask[1, [0, 4]] = 1
    observation_model = nmo.observation_models.PoissonObservations(jnp.exp)
    regularizer = nmo.regularizer.UnRegularized("GradientDescent", {})
    model = nmo.glm.PopulationGLM(observation_model, regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate, mask


@pytest.fixture
def example_data_prox_operator():
    n_features = 4

    params = (
        jnp.ones((n_features)),
        jnp.zeros(
            1,
        ),
    )
    regularizer_strength = 0.1
    mask = jnp.array([[1, 0, 1, 0], [0, 1, 0, 1]]).astype(float)
    scaling = 0.5

    return params, regularizer_strength, mask, scaling


@pytest.fixture
def example_data_prox_operator_multineuron():
    n_features = 4
    n_neurons = 3

    params = (
        jnp.ones((n_features, n_neurons)),
        jnp.zeros(
            n_neurons,
        ),
    )
    regularizer_strength = 0.1
    mask = jnp.array([[1, 0, 1, 0], [0, 1, 0, 1]], dtype=jnp.float32)
    scaling = 0.5

    return params, regularizer_strength, mask, scaling


@pytest.fixture
def poisson_observation_model():
    return nmo.observation_models.PoissonObservations(jnp.exp)


@pytest.fixture
def ridge_regularizer():
    return nmo.regularizer.Ridge()


@pytest.fixture
def lasso_regularizer():
    return nmo.regularizer.Lasso(solver_name="ProximalGradient")


@pytest.fixture
def group_lasso_2groups_5features_regularizer():
    mask = np.zeros((2, 5))
    mask[0, :2] = 1
    mask[1, 2:] = 1
    return nmo.regularizer.GroupLasso(solver_name="ProximalGradient", mask=mask)


@pytest.fixture
def mock_data():
    return jnp.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]), jnp.array([[1, 2], [3, 4]])


@pytest.fixture()
def glm_class():
    return nmo.glm.GLM


@pytest.fixture()
def population_glm_class():
    return nmo.glm.PopulationGLM


@pytest.fixture
def gammaGLM_model_instantiation():
    """Set up a Gamma GLM for testing purposes.

    This fixture initializes a Gamma GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.uniform(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.uniform(size=(5,))
    observation_model = nmo.observation_models.GammaObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer)
    rate = (jax.numpy.einsum("k,tk->t", w_true, X) + b_true) ** -1
    theta = 3
    k = rate / theta
    model.scale_ = theta
    return X, np.random.gamma(k, scale=theta), model, (w_true, b_true), rate


@pytest.fixture
def gammaGLM_model_instantiation_pytree(gammaGLM_model_instantiation):
    """Set up a Gamma GLM for testing purposes.

    This fixture initializes a Gamma GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = gammaGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(model.observation_model, model.regularizer)
    return X_tree, spikes, model_tree, true_params_tree, rate


@pytest.fixture()
def gamma_population_GLM_model():
    np.random.seed(123)
    X = np.random.uniform(size=(500, 5))
    b_true = 0.5 * np.ones((3,))
    w_true = np.random.uniform(size=(5, 3))
    observation_model = nmo.observation_models.GammaObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model, regularizer=regularizer
    )
    rate = 1 / (jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    theta = 3
    model.scale_ = theta
    y = jax.random.gamma(jax.random.PRNGKey(123), rate / theta) * theta
    return X, y, model, (w_true, b_true), rate


@pytest.fixture()
def gamma_population_GLM_model_pytree(gamma_population_GLM_model):
    X, spikes, model, true_params, rate = gamma_population_GLM_model
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model, regularizer=model.regularizer
    )
    return X_tree, spikes, model_tree, true_params_tree, rate


@pytest.fixture
def regr_data():
    np.random.seed(123)
    # define inputs and coeff
    n_samples, n_features = 50, 3
    X = np.random.normal(size=(n_samples, n_features))
    coef = np.random.normal(size=(n_features))
    # set y according to lin reg eqn
    y = X.dot(coef) + 0.1 * np.random.normal(size=(n_samples,))
    return X, y, coef


@pytest.fixture
def linear_regression(regr_data):
    X, y, coef = regr_data
    # solve least-squares
    ols, _, _, _ = np.linalg.lstsq(X, y, rcond=-1)

    # set the loss
    def loss(params, X, y):
        return jnp.power(y - jnp.dot(X, params), 2).mean()

    return X, y, coef, ols, loss


@pytest.fixture
def ridge_regression(regr_data):
    X, y, coef = regr_data

    # solve least-squares
    yagu = np.hstack((y, np.zeros_like(coef)))
    Xagu = np.vstack((X, np.sqrt(0.5) * np.eye(coef.shape[0])))
    ridge, _, _, _ = np.linalg.lstsq(Xagu, yagu, rcond=-1)

    # set the loss
    def loss(params, XX, yy):
        return (
            jnp.power(yy - jnp.dot(XX, params), 2).sum()
            + 0.5 * jnp.power(params, 2).sum()
        )

    return X, y, coef, ridge, loss


@pytest.fixture
def linear_regression_tree(linear_regression):
    X, y, coef, ols, loss = linear_regression
    X_tree = dict(input_1=X[..., :2], input_2=X[..., 2:])
    coef_tree = dict(input_1=coef[:2], input_2=coef[2:])
    ols_tree = dict(input_1=ols[:2], input_2=ols[2:])

    nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, X_tree, coef_tree)

    def loss_tree(params, XX, yy):
        pred = nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, XX, params)
        return jnp.power(yy - pred, 2).sum()

    return X_tree, y, coef_tree, ols_tree, loss_tree


@pytest.fixture()
def ridge_regression_tree(ridge_regression):
    X, y, coef, ridge, loss = ridge_regression
    X_tree = dict(input_1=X[..., :2], input_2=X[..., 2:])
    coef_tree = dict(input_1=coef[:2], input_2=coef[2:])
    ridge_tree = dict(input_1=ridge[:2], input_2=ridge[2:])

    def loss_tree(params, XX, yy):
        pred = nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, XX, params)
        norm = (
            0.5
            * nmo.tree_utils.pytree_map_and_reduce(
                lambda x: jnp.power(x, 2).sum(), sum, params
            ).sum()
        )
        return jnp.power(yy - pred, 2).sum() + norm

    return X_tree, y, coef_tree, ridge_tree, loss_tree


@pytest.fixture()
def example_X_y_high_firing_rates():
    """Example that used failed with NeMoS original initialization."""
    np.random.seed(123)

    n_features = 18
    n_neurons = 60
    n_samples = 500

    # random design array. Shape (n_time_points, n_features).
    X = 0.5 * np.random.normal(size=(n_samples, n_features))

    # log-rates & weights
    b_true = np.random.uniform(size=(n_neurons,)) * 3  # baseline rates
    w_true = np.random.uniform(size=(n_features, n_neurons)) * 0.1  # real weights:

    # generate counts (spikes will be (n_samples, n_features)
    rate = jnp.exp(jnp.dot(X, w_true) + b_true)
    spikes = np.random.poisson(rate)
    return X, spikes



import pickle
from contextlib import nullcontext as does_not_raise
from copy import deepcopy

import numpy as np
import pytest
from conftest import CombinedBasis, list_all_basis_classes
from sklearn.base import clone as sk_clone
from sklearn.pipeline import Pipeline

import nemos as nmo
from nemos import basis
from nemos._inspect_utils import get_subclass_methods, list_abstract_methods
from nemos.basis import AdditiveBasis, HistoryConv, IdentityEval, MultiplicativeBasis


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_has_the_same_public_attributes_as_basis(
    basis_cls, basis_class_specific_params
):
    n_basis_funcs = 5
    bas = CombinedBasis().instantiate_basis(
        n_basis_funcs, basis_cls, basis_class_specific_params, window_size=10
    )

    public_attrs_basis = {attr for attr in dir(bas) if not attr.startswith("_")}
    public_attrs_transformerbasis = {
        attr
        for attr in dir(
            bas.set_input_shape(*([1] * bas._n_input_dimensionality)).to_transformer()
        )
        if not attr.startswith("_")
    }

    assert public_attrs_transformerbasis - public_attrs_basis == {
        "fit",
        "fit_transform",
        "transform",
        "basis",
    }

    assert public_attrs_basis - public_attrs_transformerbasis == {"to_transformer"}


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv") + list_all_basis_classes("Eval"),
)
def test_to_transformer_and_constructor_are_equivalent(
    basis_cls, basis_class_specific_params
):
    n_basis_funcs = 5
    bas = CombinedBasis().instantiate_basis(
        n_basis_funcs, basis_cls, basis_class_specific_params, window_size=10
    )
    bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    trans_bas_a = bas.to_transformer()
    trans_bas_b = basis.TransformerBasis(bas)

    # they both just have a _basis
    assert (
        list(trans_bas_a.__dict__.keys())
        == list(trans_bas_b.__dict__.keys())
        == ["basis", "_wrapped_methods"]
    )
    # and those bases are the same
    assert np.all(
        trans_bas_a.basis.__dict__.pop("_decay_rates", 1)
        == trans_bas_b.basis.__dict__.pop("_decay_rates", 1)
    )

    # extract the wrapped func for these methods
    wrapped_methods_a = {}
    for method in trans_bas_a._chainable_methods:
        out = trans_bas_a.basis.__dict__.pop(method, False)
        val = out if out is False else out.__func__.__qualname__
        wrapped_methods_a.update({method: val})

    wrapped_methods_b = {}
    for method in trans_bas_b._chainable_methods:
        out = trans_bas_b.basis.__dict__.pop(method, False)
        val = out if out is False else out.__func__.__qualname__
        wrapped_methods_b.update({method: val})

    assert wrapped_methods_a == wrapped_methods_b
    assert trans_bas_a.basis.__dict__ == trans_bas_b.basis.__dict__


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_basis_to_transformer_makes_a_copy(basis_cls, basis_class_specific_params):

    if basis_cls in [nmo.basis.IdentityEval, nmo.basis.HistoryConv]:
        return

    bas_a = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas_a = bas_a.set_input_shape(
        *([1] * bas_a._n_input_dimensionality)
    ).to_transformer()

    # changing an attribute in bas should not change trans_bas
    if basis_cls in [basis.AdditiveBasis, basis.MultiplicativeBasis]:
        bas_a.basis1.n_basis_funcs = 10
        assert trans_bas_a.basis.basis1.n_basis_funcs == 5

        # changing an attribute in the transformer basis should not change the original
        bas_b = CombinedBasis().instantiate_basis(
            5, basis_cls, basis_class_specific_params, window_size=10
        )
        bas_b.set_input_shape(*([1] * bas_b._n_input_dimensionality))
        trans_bas_b = bas_b.to_transformer()
        trans_bas_b.basis.basis1.n_basis_funcs = 100
        assert bas_b.basis1.n_basis_funcs == 5
    else:
        bas_a.n_basis_funcs = 10
        assert trans_bas_a.n_basis_funcs == 5

        # changing an attribute in the transformer basis should not change the original
        bas_b = CombinedBasis().instantiate_basis(
            5, basis_cls, basis_class_specific_params, window_size=10
        )
        trans_bas_b = bas_b.set_input_shape(
            *([1] * bas_b._n_input_dimensionality)
        ).to_transformer()
        trans_bas_b.n_basis_funcs = 100
        assert bas_b.n_basis_funcs == 5


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize("n_basis_funcs", [5, 10, 20])
def test_transformerbasis_getattr(
    basis_cls, n_basis_funcs, basis_class_specific_params
):
    bas = CombinedBasis().instantiate_basis(
        n_basis_funcs, basis_cls, basis_class_specific_params, window_size=30
    )
    trans_basis = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )
    if basis_cls in [basis.AdditiveBasis, basis.MultiplicativeBasis]:
        for basi in [getattr(trans_basis.basis, attr) for attr in ("basis1", "basis2")]:
            assert basi.n_basis_funcs == bas.basis1.n_basis_funcs
    else:
        assert trans_basis.n_basis_funcs == bas.n_basis_funcs


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv") + list_all_basis_classes("Eval"),
)
@pytest.mark.parametrize("n_basis_funcs_init", [5])
@pytest.mark.parametrize("n_basis_funcs_new", [6, 10, 20])
def test_transformerbasis_set_params(
    basis_cls, n_basis_funcs_init, n_basis_funcs_new, basis_class_specific_params
):
    if basis_cls in [nmo.basis.IdentityEval]:
        return  # no settable params

    bas = CombinedBasis().instantiate_basis(
        n_basis_funcs_init, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_basis = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )
    if not isinstance(bas, HistoryConv):
        trans_basis.set_params(n_basis_funcs=n_basis_funcs_new)
        assert trans_basis.n_basis_funcs == n_basis_funcs_new
        assert trans_basis.basis.n_basis_funcs == n_basis_funcs_new
    else:
        trans_basis.set_params(window_size=n_basis_funcs_new)
        assert trans_basis.window_size == n_basis_funcs_new
        assert trans_basis.basis.window_size == n_basis_funcs_new


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv") + list_all_basis_classes("Eval"),
)
def test_transformerbasis_setattr_basis(basis_cls, basis_class_specific_params):

    # setting the _basis attribute should change it
    bas = CombinedBasis().instantiate_basis(
        10, basis_cls, basis_class_specific_params, window_size=30
    )
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )

    bas = CombinedBasis().instantiate_basis(
        20, basis_cls, basis_class_specific_params, window_size=30
    )
    nbas = deepcopy(bas.n_basis_funcs)

    trans_bas.basis = bas.set_input_shape(*([1] * bas._n_input_dimensionality))

    assert trans_bas.n_basis_funcs == nbas
    assert trans_bas.basis.n_basis_funcs == nbas
    assert isinstance(trans_bas.basis, basis_cls)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv") + list_all_basis_classes("Eval"),
)
def test_transformerbasis_setattr_basis_attribute(
    basis_cls, basis_class_specific_params
):
    if basis_cls in [nmo.basis.IdentityEval]:
        return
    # setting an attribute that is an attribute of the underlying _basis
    # should propagate setting it on _basis itself
    bas = CombinedBasis().instantiate_basis(
        10, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )
    if basis_cls is nmo.basis.HistoryConv:
        trans_bas.window_size = 20
    else:
        trans_bas.n_basis_funcs = 20
    assert trans_bas.n_basis_funcs == 20
    assert trans_bas.basis.n_basis_funcs == 20
    assert isinstance(trans_bas.basis, basis_cls)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv") + list_all_basis_classes("Eval"),
)
def test_transformerbasis_copy_basis_on_construct(
    basis_cls, basis_class_specific_params
):
    if basis_cls in [nmo.basis.IdentityEval]:
        return

    # modifying the transformerbasis's attributes shouldn't
    # touch the original basis that was used to create it
    orig_bas = CombinedBasis().instantiate_basis(
        10, basis_cls, basis_class_specific_params, window_size=10
    )
    nbas = deepcopy(orig_bas.n_basis_funcs)
    orig_bas = orig_bas.set_input_shape(*([1] * orig_bas._n_input_dimensionality))
    trans_bas = basis.TransformerBasis(orig_bas)
    attr_name = "window_size" if basis_cls is HistoryConv else "n_basis_funcs"
    setattr(trans_bas, attr_name, 20)

    assert orig_bas.n_basis_funcs == nbas
    assert trans_bas.basis.n_basis_funcs == 20
    assert trans_bas.basis.n_basis_funcs == 20
    assert isinstance(trans_bas.basis, basis_cls)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformerbasis_setattr_illegal_attribute(
    basis_cls, basis_class_specific_params
):
    # changing an attribute that is not _basis or an attribute of _basis
    # is not allowed
    bas = CombinedBasis().instantiate_basis(
        10, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )

    with pytest.raises(
        ValueError,
        match="Only setting basis or existing attributes of basis is allowed.",
    ):
        trans_bas.random_attr = "random value"


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformerbasis_addition(basis_cls, basis_class_specific_params):

    if basis_cls in [nmo.basis.IdentityEval, nmo.basis.HistoryConv]:
        return

    n_basis_funcs_a = 5
    n_basis_funcs_b = n_basis_funcs_a * 2
    bas_a = CombinedBasis().instantiate_basis(
        n_basis_funcs_a, basis_cls, basis_class_specific_params, window_size=10
    )
    bas_a.set_input_shape(*([1] * bas_a._n_input_dimensionality))
    bas_b = CombinedBasis().instantiate_basis(
        n_basis_funcs_b, basis_cls, basis_class_specific_params, window_size=10
    )
    bas_b.set_input_shape(*([1] * bas_b._n_input_dimensionality))
    trans_bas_a = basis.TransformerBasis(bas_a)
    trans_bas_b = basis.TransformerBasis(bas_b)
    trans_bas_sum = trans_bas_a + trans_bas_b
    assert isinstance(trans_bas_sum, basis.TransformerBasis)
    assert isinstance(trans_bas_sum.basis, basis.AdditiveBasis)
    assert (
        trans_bas_sum.n_basis_funcs
        == trans_bas_a.n_basis_funcs + trans_bas_b.n_basis_funcs
    )
    assert (
        trans_bas_sum._n_input_dimensionality
        == trans_bas_a._n_input_dimensionality + trans_bas_b._n_input_dimensionality
    )
    if basis_cls not in [basis.AdditiveBasis, basis.MultiplicativeBasis]:
        assert trans_bas_sum.basis1.n_basis_funcs == n_basis_funcs_a
        assert trans_bas_sum.basis2.n_basis_funcs == n_basis_funcs_b


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformerbasis_multiplication(basis_cls, basis_class_specific_params):

    n_basis_funcs_a = 5
    n_basis_funcs_b = n_basis_funcs_a * 2
    bas1 = CombinedBasis().instantiate_basis(
        n_basis_funcs_a, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas_a = basis.TransformerBasis(
        bas1.set_input_shape(*([1] * bas1._n_input_dimensionality))
    )
    bas2 = CombinedBasis().instantiate_basis(
        n_basis_funcs_b, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas_b = basis.TransformerBasis(
        bas2.set_input_shape(*([1] * bas2._n_input_dimensionality))
    )
    trans_bas_prod = trans_bas_a * trans_bas_b
    assert isinstance(trans_bas_prod, basis.TransformerBasis)
    assert isinstance(trans_bas_prod.basis, basis.MultiplicativeBasis)
    assert (
        trans_bas_prod.n_basis_funcs
        == trans_bas_a.n_basis_funcs * trans_bas_b.n_basis_funcs
    )
    assert (
        trans_bas_prod._n_input_dimensionality
        == trans_bas_a._n_input_dimensionality + trans_bas_b._n_input_dimensionality
    )
    if basis_cls not in [basis.AdditiveBasis, basis.MultiplicativeBasis]:
        assert trans_bas_prod.basis1.n_basis_funcs == bas1.n_basis_funcs
        assert trans_bas_prod.basis2.n_basis_funcs == bas2.n_basis_funcs


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize(
    "exponent, error_type, error_message",
    [
        (2, does_not_raise, None),
        (5, does_not_raise, None),
        (0.5, TypeError, "Exponent should be an integer"),
        (-1, ValueError, "Exponent should be a non-negative integer"),
    ],
)
def test_transformerbasis_exponentiation(
    basis_cls, exponent: int, error_type, error_message, basis_class_specific_params
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )

    if not isinstance(exponent, int):
        with pytest.raises(error_type, match=error_message):
            trans_bas_exp = trans_bas**exponent
            assert isinstance(trans_bas_exp, basis.TransformerBasis)
            assert isinstance(trans_bas_exp.basis, basis.MultiplicativeBasis)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformerbasis_dir(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )
    for attr_name in (
        "fit",
        "transform",
        "fit_transform",
        "n_basis_funcs",
        "mode",
        "window_size",
    ):
        if (
            attr_name == "window_size"
            and "Conv" not in trans_bas.basis.__class__.__name__
        ):
            continue
        assert attr_name in dir(trans_bas)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes("Conv"),
)
def test_transformerbasis_sk_clone_kernel_noned(basis_cls, basis_class_specific_params):
    orig_bas = CombinedBasis().instantiate_basis(
        10, basis_cls, basis_class_specific_params, window_size=20
    )
    orig_bas.set_input_shape(*([1] * orig_bas._n_input_dimensionality))
    trans_bas = basis.TransformerBasis(orig_bas)

    # kernel should be saved in the object after fit
    trans_bas.fit(np.random.randn(100, 1))
    assert isinstance(trans_bas.kernel_, np.ndarray)

    # cloning should set kernel_ to None
    trans_bas_clone = sk_clone(trans_bas)

    # the original object should still have kernel_
    assert isinstance(trans_bas.kernel_, np.ndarray)
    # but the clone should not have one
    assert trans_bas_clone.kernel_ is None


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize("n_basis_funcs", [5])
def test_transformerbasis_pickle(
    tmpdir, basis_cls, n_basis_funcs, basis_class_specific_params
):

    bas = CombinedBasis().instantiate_basis(
        n_basis_funcs, basis_cls, basis_class_specific_params, window_size=10
    )
    # the test that tries cross-validation with n_jobs = 2 already should test this
    trans_bas = basis.TransformerBasis(
        bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    )
    filepath = tmpdir / "transformerbasis.pickle"
    with open(filepath, "wb") as f:
        pickle.dump(trans_bas, f)
    with open(filepath, "rb") as f:
        trans_bas2 = pickle.load(f)

    assert isinstance(trans_bas2, basis.TransformerBasis)
    if basis_cls in [basis.AdditiveBasis, basis.MultiplicativeBasis]:
        for basi in [getattr(trans_bas2.basis, attr) for attr in ("basis1", "basis2")]:

            assert basi.n_basis_funcs == bas.basis1.n_basis_funcs
    else:
        assert trans_bas2.n_basis_funcs == bas.n_basis_funcs


@pytest.mark.parametrize(
    "set_input, expectation",
    [
        (True, does_not_raise()),
        (
            False,
            pytest.raises(
                RuntimeError,
                match="Cannot apply TransformerBasis: the provided basis has no defined input shape",
            ),
        ),
    ],
)
@pytest.mark.parametrize(
    "inp", [np.ones((10,)), np.ones((10, 1)), np.ones((10, 2)), np.ones((10, 2, 3))]
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_to_transformer_and_set_input(
    basis_cls, inp, set_input, expectation, basis_class_specific_params
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    if set_input:
        bas.set_input_shape(*([inp] * bas._n_input_dimensionality))
    trans = bas.to_transformer()
    with expectation:
        X = np.concatenate(
            [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
        )
        trans.fit(X)


@pytest.mark.parametrize(
    "inp, expectation",
    [
        (np.ones((10,)), pytest.raises(ValueError, match="X must be 2-")),
        (np.ones((10, 1)), does_not_raise()),
        (np.ones((10, 2)), does_not_raise()),
        (np.ones((10, 2, 3)), pytest.raises(ValueError, match="X must be 2-")),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_fit(basis_cls, inp, basis_class_specific_params, expectation):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.concatenate(
        [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
    )
    transformer.fit(X)
    if "Conv" in basis_cls.__name__:
        assert transformer.kernel_ is not None

    # try and pass segmented time series
    if isinstance(bas, (basis.AdditiveBasis, basis.MultiplicativeBasis)):
        if inp.ndim == 2:
            expectation = pytest.raises(ValueError, match="Input mismatch: expected ")

    with expectation:
        transformer.fit(*([inp] * bas._n_input_dimensionality))


@pytest.mark.parametrize(
    "inp",
    [
        np.ones((10, 1)),
        np.ones((10, 2)),
    ],
)
@pytest.mark.parametrize(
    "delta_input, expectation",
    [
        (0, does_not_raise()),
        (1, pytest.raises(ValueError, match="Input mismatch: expected ")),
        (-1, pytest.raises(ValueError, match="Input mismatch: expected ")),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_fit_input_shape_mismatch(
    basis_cls, delta_input, inp, basis_class_specific_params, expectation
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.random.randn(10, int(sum(bas._input_shape_product) + delta_input))
    with expectation:
        transformer.fit(X)


@pytest.mark.parametrize(
    "inp",
    [
        np.random.randn(
            10,
        ),
        np.random.randn(10, 1),
        np.random.randn(10, 2),
        np.random.randn(10, 2, 3),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_transform(basis_cls, inp, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.concatenate(
        [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
    )
    transformer.fit(X)

    out = transformer.transform(X)
    out2 = bas.compute_features(*([inp] * bas._n_input_dimensionality))

    assert np.array_equal(out, out2, equal_nan=True)


@pytest.mark.parametrize(
    "inp",
    [
        np.random.randn(
            10,
        ),
        np.random.randn(10, 1),
        np.random.randn(10, 2),
        np.random.randn(10, 2, 3),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_fit_transform(basis_cls, inp, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.concatenate(
        [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
    )

    out = transformer.fit_transform(X)
    out2 = bas.compute_features(*([inp] * bas._n_input_dimensionality))

    assert np.array_equal(out, out2, equal_nan=True)


@pytest.mark.parametrize(
    "inp",
    [
        np.ones((10, 1)),
        np.ones((10, 2)),
    ],
)
@pytest.mark.parametrize(
    "delta_input, expectation",
    [
        (0, does_not_raise()),
        (1, pytest.raises(ValueError, match="Input mismatch: expected ")),
        (-1, pytest.raises(ValueError, match="Input mismatch: expected ")),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_fit_transform_input_shape_mismatch(
    basis_cls, delta_input, inp, basis_class_specific_params, expectation
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.random.randn(10, int(sum(bas._input_shape_product) + delta_input))
    with expectation:
        transformer.fit_transform(X)


@pytest.mark.parametrize(
    "inp, expectation",
    [
        (np.ones((10,)), pytest.raises(ValueError, match="X must be 2-")),
        (np.ones((10, 1)), does_not_raise()),
        (np.ones((10, 2)), does_not_raise()),
        (np.ones((10, 2, 3)), pytest.raises(ValueError, match="X must be 2-")),
    ],
)
@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_transformer_fit_transform_input_struct(
    basis_cls, inp, basis_class_specific_params, expectation
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    X = np.concatenate(
        [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
    )
    transformer.fit_transform(X)

    if "Conv" in basis_cls.__name__:
        assert transformer.kernel_ is not None

    # try and pass a tuple of time series
    if (
        isinstance(bas, (basis.AdditiveBasis, basis.MultiplicativeBasis))
        and inp.ndim != 2
    ):
        expectation = pytest.raises(ValueError, match="X must be 2-")
    elif (
        isinstance(bas, (basis.AdditiveBasis, basis.MultiplicativeBasis))
        and inp.ndim == 2
    ):
        expectation = pytest.raises(ValueError, match="Input mismatch: expected")
    with expectation:
        transformer.fit(*([inp] * bas._n_input_dimensionality))


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize(
    "inp",
    [
        0.1
        * np.random.randn(
            100,
        ),
        0.1 * np.random.randn(100, 1),
        0.1 * np.random.randn(100, 2),
        0.1 * np.random.randn(100, 1, 2),
    ],
)
def test_transformer_in_pipeline(basis_cls, inp, basis_class_specific_params):

    if basis_cls is IdentityEval:
        return

    cv_attr = "n_basis_funcs" if basis_cls is not HistoryConv else "window_size"
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=5
    )
    transformer = bas.set_input_shape(
        *([inp] * bas._n_input_dimensionality)
    ).to_transformer()
    # fit outside pipeline
    X = bas.compute_features(*([inp] * bas._n_input_dimensionality))
    log_mu = X.dot(0.005 * np.ones(X.shape[1]))
    y = np.full(X.shape[0], 0)
    y[~np.isnan(log_mu)] = np.random.poisson(
        np.exp(log_mu[~np.isnan(log_mu)] - np.nanmean(log_mu))
    )
    model = nmo.glm.GLM(regularizer="Ridge", regularizer_strength=0.001).fit(X, y)

    # pipeline
    pipe = Pipeline(
        [
            ("bas", transformer),
            ("glm", nmo.glm.GLM(regularizer="Ridge", regularizer_strength=0.001)),
        ]
    )
    x = np.concatenate(
        [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
    )
    pipe.fit(x, y)
    np.testing.assert_allclose(pipe["glm"].coef_, model.coef_)

    set_param_dict = {f"bas__basis2__{cv_attr}": 4}
    # set basis & refit
    if isinstance(bas, (basis.AdditiveBasis, basis.MultiplicativeBasis)):
        pipe.set_params(**set_param_dict)
        assert (
            bas.basis2.n_basis_funcs == 5
        )  # make sure that the change did not affect bas
        set_param_dict_outside = {f"basis2__{cv_attr}": 4}
        X = bas.set_params(**set_param_dict_outside).compute_features(
            *([inp] * bas._n_input_dimensionality)
        )
    else:
        set_param_dict = {f"bas__{cv_attr}": 4}
        pipe.set_params(**set_param_dict)
        assert bas.n_basis_funcs == 5  # make sure that the change did not affect bas
        set_param_dict_outside = {f"{cv_attr}": 4}
        X = bas.set_params(**set_param_dict_outside).compute_features(
            *([inp] * bas._n_input_dimensionality)
        )
    pipe.fit(x, y)
    model.fit(X, y)
    np.testing.assert_allclose(pipe["glm"].coef_, model.coef_)


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_initialization(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.to_transformer()
    with pytest.raises(RuntimeError, match="Cannot apply TransformerBasis"):
        transformer.fit(np.ones((100,)))

    with pytest.raises(RuntimeError, match="Cannot apply TransformerBasis"):
        transformer.transform(np.ones((100,)))

    with pytest.raises(RuntimeError, match="Cannot apply TransformerBasis"):
        transformer.fit_transform(np.ones((100,)))


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_basis_setter(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )

    bas2 = CombinedBasis().instantiate_basis(
        7, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.to_transformer()
    transformer.basis = bas2
    assert transformer.basis.n_basis_funcs == bas2.n_basis_funcs


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_getstate(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.to_transformer()
    state = transformer.__getstate__()
    assert {"basis": transformer.basis} == state


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_eetstate(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    bas2 = CombinedBasis().instantiate_basis(
        7, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.to_transformer()
    state = {"basis": bas2}
    transformer.__setstate__(state)
    assert transformer.basis == bas2


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_to_transformer_not_an_attribute_of_transformer_basis(
    basis_cls, basis_class_specific_params
):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    bas = bas.to_transformer()
    assert "to_transformer" not in bas.__dir__()

    with pytest.raises(
        AttributeError,
        match="'TransformerBasis' object has no attribute 'to_transformer'",
    ):
        bas.to_transformer()


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
def test_dir_transformer(basis_cls, basis_class_specific_params):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    transformer = bas.to_transformer()
    lst = transformer.__dir__()
    dict_abst_method = list_abstract_methods(nmo.basis._basis.Basis)

    # check it finds all abc basis methods
    for meth in dict_abst_method:
        assert meth[0] in lst

    # check all reimplemented methods
    dict_reimplemented_method = get_subclass_methods(basis_cls)
    for meth in dict_reimplemented_method:
        if meth[0] == "to_transformer":
            continue
        assert meth[0] in lst

    # check that it is a trnasformer
    for meth in ["fit", "transform", "fit_transform"]:
        assert meth in lst


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize(
    "inp, expectation",
    [
        (
            np.random.randn(10, 2),
            pytest.raises(ValueError, match=r"Input mismatch: expected \d inputs"),
        ),
        (
            np.random.randn(10, 3, 1),
            pytest.raises(ValueError, match="X must be 2-dimensional"),
        ),
        (
            {1: np.random.randn(10, 3)},
            pytest.raises(ValueError, match="The input must be a 2-dimensional array"),
        ),
        (np.random.randn(10, 3), does_not_raise()),
    ],
)
@pytest.mark.parametrize("method", ["fit", "transform", "fit_transform"])
def test_check_input(inp, expectation, basis_cls, basis_class_specific_params, method):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    # set kernels
    bas._set_input_independent_states()
    # set input shape
    transformer = bas.to_transformer().set_input_shape(
        *([3] * bas._n_input_dimensionality)
    )
    if isinstance(bas, (AdditiveBasis, MultiplicativeBasis)):
        if hasattr(inp, "ndim"):
            ndim = inp.ndim
            inp = np.concatenate(
                [inp.reshape(inp.shape[0], -1)] * bas._n_input_dimensionality, axis=1
            )
            if ndim == 3:
                inp = inp[..., np.newaxis]

    meth = getattr(transformer, method)

    with expectation:
        meth(inp)
        with pytest.raises(ValueError, match="X and y must have the same"):
            meth(inp, np.ones(11))


@pytest.mark.parametrize(
    "basis_cls",
    list_all_basis_classes(),
)
@pytest.mark.parametrize(
    "expected_out",
    [
        {
            basis.BSplineConv: "Transformer(BSplineConv(n_basis_funcs=5, window_size=10, order=4))",
            basis.BSplineEval: "Transformer(BSplineEval(n_basis_funcs=5, order=4))",
            basis.CyclicBSplineConv: "Transformer(CyclicBSplineConv(n_basis_funcs=5, window_size=10, order=4))",
            basis.CyclicBSplineEval: "Transformer(CyclicBSplineEval(n_basis_funcs=5, order=4))",
            basis.HistoryConv: "Transformer(HistoryConv(window_size=10))",
            basis.IdentityEval: "Transformer(IdentityEval())",
            basis.MSplineConv: "Transformer(MSplineConv(n_basis_funcs=5, window_size=10, order=4))",
            basis.MSplineEval: "Transformer(MSplineEval(n_basis_funcs=5, order=4))",
            basis.OrthExponentialConv: "Transformer(OrthExponentialConv(n_basis_funcs=5, window_size=10))",
            basis.OrthExponentialEval: "Transformer(OrthExponentialEval(n_basis_funcs=5))",
            basis.RaisedCosineLinearConv: "Transformer(RaisedCosineLinearConv(n_basis_funcs=5, window_size=10, width=2.0))",
            basis.RaisedCosineLinearEval: "Transformer(RaisedCosineLinearEval(n_basis_funcs=5, width=2.0))",
            basis.RaisedCosineLogConv: "Transformer(RaisedCosineLogConv(n_basis_funcs=5, window_size=10, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True))",
            basis.RaisedCosineLogEval: "Transformer(RaisedCosineLogEval(n_basis_funcs=5, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True))",
            basis.AdditiveBasis: "Transformer(AdditiveBasis(\n    basis1=MSplineEval(n_basis_funcs=5, order=4),\n    basis2=RaisedCosineLinearConv(n_basis_funcs=5, window_size=10, width=2.0),\n))",
            basis.MultiplicativeBasis: "Transformer(MultiplicativeBasis(\n    basis1=MSplineEval(n_basis_funcs=5, order=4),\n    basis2=RaisedCosineLinearConv(n_basis_funcs=5, window_size=10, width=2.0),\n))",
        }
    ],
)
def test_repr_out(basis_cls, basis_class_specific_params, expected_out):
    bas = CombinedBasis().instantiate_basis(
        5, basis_cls, basis_class_specific_params, window_size=10
    )
    bas = bas.set_input_shape(*([10] * bas._n_input_dimensionality)).to_transformer()
    out = expected_out.get(basis_cls, "")
    if out == "":
        raise ValueError(f"Missing test case for {basis_cls}!")
    assert repr(bas) == out



from typing import Callable

import jax
import jax.numpy as jnp
from numpy.typing import ArrayLike
from scipy.optimize import root_scalar

# dictionary of known inverse link functions.
INVERSE_FUNCS = {
    jnp.exp: jnp.log,
    jax.nn.softplus: lambda x: jnp.log(jnp.exp(x) - 1.0),
}


def scalar_root_find_elementwise(
    func: Callable, args: ArrayLike, x0: ArrayLike
) -> jnp.ndarray:
    """
    Find roots of a scalar function.

    This can be used as an attempt to find a numerical inverse of an unknown link function of a GLM; typically,
    this numerical inverse, is used to set the initial intercept to match the mean firing rate of the model.

    Parameters
    ----------
    func:
        A callable, which typically will be `inv_link_func(x) - jnp.mean(spikes)`.
    args:
        List of additional arguments passed to the function.
    x0:
        Initial values for the root-finding algorithm.

    Returns
    -------
    :
        An array containing the roots of each f(x) = func(x, args[k]), for k in 1,..., len(args).

    Raises
    ------
    ValueError:
        If any of the optimization is not successful.
    """
    opts = [root_scalar(func, arg, x0=x, method="secant") for arg, x in zip(args, x0)]

    if not all(jnp.abs(func(opt.root, args[i])) < 10**-4 for i, opt in enumerate(opts)):
        raise ValueError(
            "Could not set the initial intercept as the inverse of the firing rate for "
            "the provided link function. "
            "Please, provide initial parameters instead!"
        )

    return jnp.array([opt.root for opt in opts])


def initialize_intercept_matching_mean_rate(
    inverse_link_function: Callable, y: jnp.ndarray
) -> jnp.ndarray:
    """
    Compute the initial intercept term for a regression models.

    This method compute an initial intercept term for a regression models such that the baseline activity
    matches the mean activity of each neuron, assuming that the model coefficients are initialized to zero.


    Parameters
    ----------
    inverse_link_function:
        The inverse link function of the model, linking the mean to the linear combination of the covariates in
        a GLM.
    y:
        The neural activity, shape either (num_sample,) for single variable regressors as `GLM`
         or (n_sample, n_neurons) for multi-variable regressors, such as `PopulaitonGLM`.

    Returns
    -------
    :
        The initial intercept term, shape (n_neurons,).

    """

    # return inverse if analytical solution is available
    analytical_inv = INVERSE_FUNCS.get(inverse_link_function, None)

    means = jnp.atleast_1d(jnp.nanmean(y, axis=0))
    if analytical_inv:
        out = analytical_inv(means)
        if jnp.any(jnp.isnan(out)):
            raise ValueError(
                "Failed to initialize the model intercept as the inverse of the firing rate for "
                "the provided link function. The mean firing rate has some non-positive values."
            )
        return out

    def func(x, mean_x):
        return inverse_link_function(x) - mean_x

    try:
        out = scalar_root_find_elementwise(func, means, means)
    except ValueError:
        raise ValueError(
            "Failed to initialize the model intercept as the inverse of the firing rate for the"
            " provided link function. Please, provide initial parameters instead!"
        )
    return out



#!/usr/bin/env python3
from collections import UserDict

import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from .tree_utils import pytree_map_and_reduce


@register_pytree_node_class
class FeaturePytree(UserDict):
    """Pytree to represent GLM features.

    This object is essentially a dictionary with strings as its keys and
    n-dimensional array-like objects as its values. FeaturePytree objects can
    only have a depth of 1, and we allow joint slicing.

    This is intended to be used with jax.tree_util.tree_map and similar functionality.
    """

    def __init__(self, **kwargs):
        self._num_time_points = None
        super().__init__(kwargs)

    @property
    def shape(self):
        # Every value is an array with the same number of time points, so
        # that's the only shared shape.
        return (self._num_time_points,)

    def __len__(self):
        # Same logic as shape
        return self._num_time_points

    @property
    def ndim(self):
        # You can slice into a FeaturePytree along one dimension (time)
        return 1

    def __setitem__(self, key, value):
        # All keys are strings
        if not isinstance(key, str):
            raise ValueError("Keys must be strings!")
        # All values are array-like and must have time points along the first
        # dimension, so double-check that's the case here.
        try:
            value.shape[0]
        except (AttributeError, IndexError):
            raise ValueError("All values must be arrays of at least 1 dimension!")
        # For the first value, we must set how time points this FeaturePytree
        # has
        if self._num_time_points is None:
            self._num_time_points = value.shape[0]
        # For all subsequent values, ensure they have the proper number of time points
        else:
            if self._num_time_points != value.shape[0]:
                raise ValueError(
                    f"All arrays must have same number of time points, {self._num_time_points}"
                )
        super().__setitem__(key, value)

    def __getitem__(self, key):
        # We can index into either the features, which have strs as their
        # keys...
        if isinstance(key, str):
            return self.data[key]
        # Or the time dimension
        else:
            return jax.tree_util.tree_map(lambda x: x[key], self)

    def __repr__(self):
        # Show the shape and data type of each array
        repr = [f"{k}: shape {v.shape}, dtype {v.dtype}" for k, v in self.data.items()]
        return "\n".join(repr)

    def __eq__(self, other):
        # if structure is different, pytree_map_and_reduce will return a ValueError
        if jax.tree_util.tree_structure(self) != jax.tree_util.tree_structure(other):
            return False
        return pytree_map_and_reduce(
            lambda x, y: jnp.array_equal(x, y), all, self, other
        )

    def tree_flatten(self):
        return jax.tree_util.tree_flatten(self.data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        try:
            return cls(**jax.tree_util.tree_unflatten(aux_data, children))
        except ValueError:
            return jax.tree_util.tree_unflatten(aux_data, children)



# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
import copy
from collections import OrderedDict
from functools import wraps
from typing import Callable, Generator, Literal, Optional, Tuple, Union

import jax
import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..base_class import Base
from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ..utils import format_repr, row_wise_kron
from ..validation import check_fraction_valid_samples
from ._basis_mixin import BasisTransformerMixin, CompositeBasisMixin


def add_docstring(method_name, cls):
    """Prepend super-class docstrings."""
    attr = getattr(cls, method_name, None)
    if attr is None:
        raise AttributeError(f"{cls.__name__} has no attribute {method_name}!")
    doc = attr.__doc__

    # Decorator to add the docstring
    def wrapper(func):
        func.__doc__ = "\n".join([doc, func.__doc__])  # Combine docstrings
        return func

    return wrapper


def check_transform_input(func: Callable) -> Callable:
    """Check input before calling basis.

    This decorator allows to raise an exception that is more readable
    when the wrong number of input is provided to _evaluate.
    """

    @wraps(func)
    def wrapper(self: Basis, *xi: ArrayLike, **kwargs) -> NDArray:
        xi = self._check_transform_input(*xi)
        return func(self, *xi, **kwargs)  # Call the basis

    return wrapper


def check_one_dimensional(func: Callable) -> Callable:
    """Check if the input is one-dimensional."""

    @wraps(func)
    def wrapper(self: Basis, *xi: NDArray, **kwargs):
        if any(x.ndim != 1 for x in xi):
            raise ValueError("Input sample must be one dimensional!")
        return func(self, *xi, **kwargs)

    return wrapper


def min_max_rescale_samples(
    sample_pts: NDArray,
    bounds: Optional[Tuple[float, float]] = None,
) -> Tuple[NDArray, NDArray]:
    """Rescale samples to [0,1].

    Parameters
    ----------
    sample_pts:
        The original samples.
    bounds:
        Sample bounds. `bounds[0]` and `bounds[1]` are mapped to 0 and 1, respectively.
        Default are `min(sample_pts), max(sample_pts)`.

    Warns
    -----
    UserWarning
        If more than 90% of the sample points contain NaNs or Infs.
    """
    sample_pts = sample_pts.astype(float)
    # if not normalize all array
    vmin = np.nanmin(sample_pts, axis=0) if bounds is None else bounds[0]
    vmax = np.nanmax(sample_pts, axis=0) if bounds is None else bounds[1]
    sample_pts[(sample_pts < vmin) | (sample_pts > vmax)] = np.nan
    sample_pts -= vmin

    scaling = np.asarray(vmax - vmin)
    # do not normalize if samples contain a single value (in which case vmax=vmin)
    scaling[scaling == 0] = 1.0
    sample_pts /= scaling

    check_fraction_valid_samples(
        sample_pts,
        err_msg="All the samples lie outside the [vmin, vmax] range.",
        warn_msg="More than 90% of the samples lie outside the [vmin, vmax] range.",
    )

    return sample_pts, scaling


class Basis(Base, abc.ABC, BasisTransformerMixin):
    """
    Abstract base class for defining basis functions for feature transformation.

    Basis functions are mathematical constructs that can represent data in alternative,
    often more compact or interpretable forms. This class provides a template for such
    transformations, with specific implementations defining the actual behavior.

    Parameters
    ----------
    mode :
        The mode of operation. 'eval' for evaluation at sample points,
        'conv' for convolutional operation.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Raises
    ------
    ValueError:
        If ``mode`` is not 'eval' or 'conv'.
    ValueError:
        If ``kwargs`` are not None and ``mode =="eval"``.
    ValueError:
        If ``kwargs`` include parameters not recognized or do not have
        default values in ``create_convolutional_predictor``.
    ValueError:
        If ``axis`` different from 0 is provided as a keyword argument (samples must always be in the first axis).
    """

    def __init__(
        self,
        mode: Literal["eval", "conv", "composite"] = "eval",
        label: Optional[str] = None,
    ) -> None:
        self._n_input_dimensionality = getattr(self, "_n_input_dimensionality", 0)

        self._mode = mode

        if label is None:
            self._label = self.__class__.__name__
        else:
            self._label = str(label)

        # specified only after inputs/input shapes are provided
        self._input_shape_product = getattr(self, "_input_shape_product", None)

        # initialize parent to None. This should not end in "_" because it is
        # a permanent property of a basis, defined at composite basis init
        self._parent = None

    @property
    def n_output_features(self) -> int | None:
        """
        Number of features returned by the basis.

        Notes
        -----
        The number of output features can be determined only when the number of inputs
        provided to the basis is known. Therefore, before the first call to ``compute_features``,
        this property will return ``None``. After that call, or after setting the input shape with
        ``set_input_shape``, ``n_output_features`` will be available.
        """
        if self._input_shape_product is not None:
            return self.n_basis_funcs * self._input_shape_product[0]
        return None

    @property
    def label(self) -> str:
        """Label for the basis."""
        return self._label

    @property
    def n_basis_funcs(self):
        """Number of basis functions."""
        return self._n_basis_funcs

    @n_basis_funcs.setter
    def n_basis_funcs(self, value):
        orig_n_basis = copy.deepcopy(getattr(self, "_n_basis_funcs", None))
        self._n_basis_funcs = value
        try:
            self._check_n_basis_min()
        except ValueError as e:
            self._n_basis_funcs = orig_n_basis
            raise e

    @property
    def mode(self):
        """Mode of operation, either ``"conv"`` or ``"eval"``."""
        return self._mode

    @check_transform_input
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Apply the basis transformation to the input data.

        This method is designed to be a high-level interface for transforming input
        data using the basis functions defined by the subclass. Depending on the basis'
        mode ('Eval' or 'Conv'), it either evaluates the basis functions at the sample
        points or performs a convolution operation between the input data and the
        basis functions.

        Parameters
        ----------
        *xi :
            Input data arrays to be transformed. The shape and content requirements
            depend on the subclass and mode of operation ('Eval' or 'Conv').

        Returns
        -------
        :
            Transformed features. In 'eval' mode, it corresponds to the basis functions
            evaluated at the input samples. In 'conv' mode, it consists of convolved
            input samples with the basis functions. The output shape varies based on
            the subclass and mode.

        Notes
        -----
        Subclasses should implement how to handle the transformation specific to their
        basis function types and operation modes.
        """
        if self._input_shape_product is None:
            self.set_input_shape(*xi)
        self._check_input_shape_consistency(*xi)
        self._set_input_independent_states()
        return self._compute_features(*xi)

    @abc.abstractmethod
    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """Convolve or evaluate the basis.

        This method is intended to be equivalent to the sklearn transformer ``transform`` method.
        As the latter, it computes the transformation assuming that all the states are already
        pre-computed by ``_fit_basis``, a method corresponding to ``fit``.

        The method differs from  transformer's ``transform`` for the structure of the input that it accepts.
        In particular, ``_compute_features`` accepts a number of different time series, one per 1D basis component,
        while ``transform`` requires all inputs to be concatenated in a single array.
        """
        pass

    @abc.abstractmethod
    def setup_basis(self, *xi: ArrayLike) -> FeatureMatrix:
        """Pre-compute all basis state variables.

        This method is intended to be equivalent to the sklearn transformer ``fit`` method.
        As the latter, it computes all the state attributes, and store it with the convention
        that the attribute name **must** end with "_", for example ``self.kernel_``,
        ``self._input_shape_``.

        The method differs from  transformer's ``fit`` for the structure of the input that it accepts.
        In particular, ``_fit_basis`` accepts a number of different time series, one per 1D basis component,
        while ``fit`` requires all inputs to be concatenated in a single array.
        """
        pass

    @abc.abstractmethod
    def _set_input_independent_states(self):
        """
        Compute all the basis states that do not depend on the input.

        An example of such state is the kernel_ for Conv bases, which can be computed
        without any input (it only depends on the basis type, the window size and the
        number of basis elements).
        """
        pass

    @abc.abstractmethod
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Set the expected input shape for the basis object.

        This method configures the shape of the input data that the basis object expects.
        ``xi`` can be specified as an integer, a tuple of integers, or derived
        from an array. The method also calculates the total number of input
        features and output features based on the number of basis functions.

        """
        pass

    @abc.abstractmethod
    def _evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Abstract method to evaluate the basis functions at given points.

        This method must be implemented by subclasses to define the specific behavior
        of the basis transformation. The implementation depends on the type of basis
        (e.g., spline, raised cosine), and it should evaluate the basis functions at
        the specified points in the domain.

        Parameters
        ----------
        *xi :
            Variable number of arguments, each representing an array of points at which
            to evaluate the basis functions. The dimensions and requirements of these
            inputs vary depending on the specific basis implementation.

        Returns
        -------
        :
            An array containing the evaluated values of the basis functions at the input
            points. The shape and structure of this array are specific to the subclass
            implementation.
        """
        pass

    def _get_samples(self, *n_samples: int) -> Generator[NDArray]:
        """Get equi-spaced samples for all the input dimensions.

        This will be used to evaluate the basis on a grid of
        points derived by the samples.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of samples in each axis of the grid.

        Returns
        -------
        :
            A generator yielding numpy arrays of linspaces from 0 to 1 of sizes specified by ``n_samples``.
        """
        # handling of defaults when evaluating on a grid
        # (i.e. when we cannot use max and min of samples)
        bounds = getattr(self, "bounds", None)
        if bounds is None:
            mn, mx = 0, 1
        else:
            mn, mx = bounds
        return (np.linspace(mn, mx, n_samples[k]) for k in range(len(n_samples)))

    @support_pynapple(conv_type="numpy")
    def _check_transform_input(
        self, *xi: ArrayLike
    ) -> Tuple[Union[NDArray, Tsd, TsdFrame]]:
        """Check transform input.

        Parameters
        ----------
        xi[0],...,xi[n] :
            The input samples, each  with shape (number of samples, ).

        Raises
        ------
        ValueError
            - If the time point number is inconsistent between inputs.
            - If the number of inputs doesn't match what the Basis object requires.
            - At least one of the samples is empty.

        """
        # check that the input is array-like (i.e., whether we can cast it to
        # numeric arrays)
        try:
            # make sure array is at least 1d (so that we succeed when only
            # passed a scalar)
            xi = tuple(np.atleast_1d(np.asarray(x, dtype=float)) for x in xi)
        # ValueError here surfaces the exception with e.g., `x=np.array["a", "b"])`
        except (TypeError, ValueError):
            raise TypeError("Input samples must be array-like of floats!")

        # check for non-empty samples
        if self._has_zero_samples(tuple(len(x) for x in xi)):
            raise ValueError("All sample provided must be non empty.")

        # checks on input and outputs
        self._check_samples_consistency(*xi)
        self._check_input_dimensionality(xi)

        return xi

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
           The number of samples.

        Returns
        -------
        X :
           Array of shape (n_samples,) containing the equi-spaced sample
           points where we've evaluated the basis.
        basis_funcs :
           Evaluated exponentially decaying basis functions, numerically
           orthogonalized, shape (n_samples, n_basis_funcs)
        """
        self._check_input_dimensionality(n_samples)

        if self._has_zero_samples(n_samples):
            raise ValueError("All sample counts provided must be greater than zero.")

        # get the samples
        sample_tuple = self._get_samples(*n_samples)
        Xs = np.meshgrid(*sample_tuple, indexing="ij")

        # evaluates the basis on a flat NDArray and reshape to match meshgrid output
        Y = self._evaluate(*tuple(grid_axis.flatten() for grid_axis in Xs)).reshape(
            (*n_samples, self.n_basis_funcs)
        )

        return *Xs, Y

    @staticmethod
    def _has_zero_samples(n_samples: Tuple[int, ...]) -> bool:
        return any([n <= 0 for n in n_samples])

    def _check_input_dimensionality(self, xi: Tuple) -> None:
        """
        Check that the number of inputs provided by the user matches the number of inputs required.

        Parameters
        ----------
        xi[0], ..., xi[n] :
            The input samples, shape (number of samples, ).

        Raises
        ------
        ValueError
            If the number of inputs doesn't match what the Basis object requires.
        """
        if len(xi) != self._n_input_dimensionality:
            raise TypeError(
                f"Input dimensionality mismatch. This basis evaluation requires {self._n_input_dimensionality} inputs, "
                f"{len(xi)} inputs provided instead."
            )

    @staticmethod
    def _check_samples_consistency(*xi: NDArray) -> None:
        """
        Check that each input provided to the Basis object has the same number of time points.

        Parameters
        ----------
        xi[0], ..., xi[n] :
            The input samples, shape (number of samples, ).

        Raises
        ------
        ValueError
            If the time point number is inconsistent between inputs.
        """
        sample_sizes = [sample.shape[0] for sample in xi]
        if any(elem != sample_sizes[0] for elem in sample_sizes):
            raise ValueError(
                "Sample size mismatch. Input elements have inconsistent sample sizes."
            )

    def __add__(self, other: Basis) -> AdditiveBasis:
        """
        Add two Basis objects together.

        Parameters
        ----------
        other
            The other Basis object to add.

        Returns
        -------
        : AdditiveBasis
            The resulting Basis object.
        """
        return AdditiveBasis(self, other)

    def __mul__(self, other: Basis) -> MultiplicativeBasis:
        """
        Multiply two Basis objects together.

        Parameters
        ----------
        other
            The other Basis object to multiply.

        Returns
        -------
        :
            The resulting Basis object.
        """
        return MultiplicativeBasis(self, other)

    def __pow__(self, exponent: int) -> MultiplicativeBasis:
        """Exponentiation of a Basis object.

        Define the power of a basis by repeatedly applying the method __multiply__.
        The exponent must be a positive integer.

        Parameters
        ----------
        exponent :
            Positive integer exponent

        Returns
        -------
        :
            The product of the basis with itself "exponent" times. Equivalent to ``self * self * ... * self``.

        Raises
        ------
        TypeError
            If the provided exponent is not an integer.
        ValueError
            If the integer is zero or negative.
        """
        if not isinstance(exponent, int):
            raise TypeError("Exponent should be an integer!")

        if exponent <= 0:
            raise ValueError("Exponent should be a non-negative integer!")

        result = self
        for _ in range(exponent - 1):
            result = result * self
        return result

    def __repr__(self):
        return format_repr(self)

    def _get_feature_slicing(
        self,
        n_inputs: Optional[tuple] = None,
        start_slice: Optional[int] = None,
        split_by_input: bool = True,
    ) -> Tuple[OrderedDict, int]:
        """
        Calculate and return the slicing for features based on the input structure.

        This method determines how to slice the features for different basis types.

        Parameters
        ----------
        n_inputs :
            The number of input basis for each component, by default it uses ``self._input_shape_product``.
        start_slice :
            The starting index for slicing, by default it starts from 0.
        split_by_input :
            Flag indicating whether to split the slicing by individual inputs or not.
            If ``False``, a single slice is generated for all inputs.

        Returns
        -------
        split_dict :
            Dictionary with keys as labels and values as slices representing
            the slicing for each input or additive component, if split_by_input equals to
            True or False respectively.
        start_slice :
            The updated starting index after slicing.

        See Also
        --------
        _get_default_slicing : Handles default slicing logic.
        _merge_slicing_dicts : Merges multiple slicing dictionaries, handling keys conflicts.
        """
        # Set default values for start_slice if not provided
        start_slice = start_slice or 0
        # Handle the default case for non-additive basis types
        # See overwritten method for recursion logic
        split_dict, start_slice = self._get_default_slicing(
            split_by_input=split_by_input, start_slice=start_slice
        )

        return split_dict, start_slice

    @staticmethod
    def _generate_unique_key(existing_dict: dict, key: str) -> str:
        """Generate a unique key if there is a conflict."""
        extra = 1
        new_key = f"{key}-{extra}"
        while new_key in existing_dict:
            extra += 1
            new_key = f"{key}-{extra}"
        return new_key

    def _get_default_slicing(
        self, split_by_input: bool, start_slice: int
    ) -> Tuple[OrderedDict, int]:
        """Handle default slicing logic."""
        if split_by_input:
            # should we remove this option?
            if self._input_shape_product[0] == 1 or isinstance(
                self, MultiplicativeBasis
            ):
                split_dict = {
                    self.label: slice(start_slice, start_slice + self.n_output_features)
                }
            else:
                split_dict = {
                    self.label: {
                        f"{i}": slice(
                            start_slice + i * self.n_basis_funcs,
                            start_slice + (i + 1) * self.n_basis_funcs,
                        )
                        for i in range(self._input_shape_product[0])
                    }
                }
        else:
            split_dict = {
                self.label: slice(start_slice, start_slice + self.n_output_features)
            }
        start_slice += self.n_output_features
        return OrderedDict(split_dict), start_slice

    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Decompose an array along a specified axis into sub-arrays based on the number of expected inputs.

        This function takes an array (e.g., a design matrix or model coefficients) and splits it along
        a designated axis.

        **How it works:**

        - If the basis expects an input shape ``(n_samples, n_inputs)``, then the feature axis length will
          be ``total_n_features = n_inputs * n_basis_funcs``. This axis is reshaped into dimensions
          ``(n_inputs, n_basis_funcs)``.

        - If the basis expects an input of shape ``(n_samples,)``, then the feature axis length will
          be ``total_n_features = n_basis_funcs``. This axis is reshaped into ``(1, n_basis_funcs)``.

        For example, if the input array ``x`` has shape ``(1, 2, total_n_features, 4, 5)``,
        then after applying this method, it will be reshaped into ``(1, 2, n_inputs, n_basis_funcs, 4, 5)``.

        The specified axis (``axis``) determines where the split occurs, and all other dimensions
        remain unchanged. See the example section below for the most common use cases.

        Parameters
        ----------
        x :
            The input array to be split, representing concatenated features, coefficients,
            or other data. The shape of ``x`` along the specified axis must match the total
            number of features generated by the basis, i.e., ``self.n_output_features``.

            **Examples:**

            - For a design matrix: ``(n_samples, total_n_features)``

            - For model coefficients: ``(total_n_features,)`` or ``(total_n_features, n_neurons)``.

        axis : int, optional
            The axis along which to split the features. Defaults to 1.
            Use ``axis=1`` for design matrices (features along columns) and ``axis=0`` for
            coefficient arrays (features along rows). All other dimensions are preserved.

        Raises
        ------
        ValueError
            If the shape of ``x`` along the specified axis does not match ``self.n_output_features``.

        Returns
        -------
        dict
            A dictionary where:

            - **Key**: Label of the basis.
            - **Value**: the array reshaped to: ``(..., n_inputs, n_basis_funcs, ...)``
        """
        if x.shape[axis] != self.n_output_features:
            raise ValueError(
                "`x.shape[axis]` does not match the expected number of features."
                f" `x.shape[axis] == {x.shape[axis]}`, while the expected number "
                f"of features is {self.n_output_features}"
            )

        # Get the slice dictionary based on predefined feature slicing
        slice_dict = self._get_feature_slicing(split_by_input=False)[0]

        # Helper function to build index tuples for each slice
        def build_index_tuple(slice_obj, axis: int, ndim: int):
            """Create an index tuple to apply a slice on the given axis."""
            index = [slice(None)] * ndim  # Initialize index for all dimensions
            index[axis] = slice_obj  # Replace the axis with the slice object
            return tuple(index)

        # Get the dict for slicing the correct axis
        index_dict = jax.tree_util.tree_map(
            lambda sl: build_index_tuple(sl, axis, x.ndim), slice_dict
        )

        # Custom leaf function to identify index tuples as leaves
        def is_leaf(val):
            # Check if it's a tuple, length matches ndim, and all elements are slice objects
            if isinstance(val, tuple) and len(val) == x.ndim:
                return all(isinstance(v, slice) for v in val)
            return False

        # Apply the slicing using the custom leaf function
        out = jax.tree_util.tree_map(lambda sl: x[sl], index_dict, is_leaf=is_leaf)

        # reshape the arrays to match input shapes
        reshaped_out = dict()
        for items, bas in zip(out.items(), self):
            key, val = items
            shape = list(val.shape)
            reshaped_out[key] = val.reshape(
                shape[:axis]
                + [*(b for sh in bas._input_shape_ for b in sh), -1]
                + shape[axis + 1 :]
            )
        return reshaped_out

    def __iter__(self):
        """Makes basis iterable. Re-implemented for additive."""
        yield self

    def __len__(self):
        return 1


class AdditiveBasis(CompositeBasisMixin, Basis):
    """
    Class representing the addition of two Basis objects.

    Parameters
    ----------
    basis1 :
        First basis object to add.
    basis2 :
        Second basis object to add.

    Examples
    --------
    >>> # Generate sample data
    >>> import numpy as np
    >>> import nemos as nmo
    >>> X = np.random.normal(size=(30, 2))

    >>> # define two basis objects and add them
    >>> basis_1 = nmo.basis.BSplineEval(10)
    >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
    >>> additive_basis = basis_1 + basis_2
    >>> additive_basis
    AdditiveBasis(
        basis1=BSplineEval(n_basis_funcs=10, order=4),
        basis2=RaisedCosineLinearEval(n_basis_funcs=15, width=2.0),
    )
    >>> # can add another basis to the AdditiveBasis object
    >>> X = np.random.normal(size=(30, 3))
    >>> basis_3 = nmo.basis.RaisedCosineLogEval(100)
    >>> additive_basis_2 = additive_basis + basis_3
    >>> additive_basis_2
    AdditiveBasis(
        basis1=AdditiveBasis(
            basis1=BSplineEval(n_basis_funcs=10, order=4),
            basis2=RaisedCosineLinearEval(n_basis_funcs=15, width=2.0),
        ),
        basis2=RaisedCosineLogEval(n_basis_funcs=100, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True),
    )
    """

    def __init__(self, basis1: Basis, basis2: Basis) -> None:
        CompositeBasisMixin.__init__(self, basis1, basis2)
        Basis.__init__(self, mode="composite")
        self._label = "(" + basis1.label + " + " + basis2.label + ")"

        # number of input arrays that the basis receives
        self._n_input_dimensionality = (
            basis1._n_input_dimensionality + basis2._n_input_dimensionality
        )

    @property
    def n_basis_funcs(self):
        """Compute the n-basis function runtime.

        This plays well with cross-validation where the number of basis function of the
        underlying bases can be changed. It must be read-only since the number of basis
        is determined by the two basis elements and the type of composition.
        """
        return self.basis1.n_basis_funcs + self.basis2.n_basis_funcs

    @property
    def n_output_features(self):
        out1 = getattr(self.basis1, "n_output_features", None)
        out2 = getattr(self.basis2, "n_output_features", None)
        if out1 is None or out2 is None:
            return None
        return out1 + out2

    @add_docstring("set_input_shape", CompositeBasisMixin)
    def set_input_shape(self, *xi: int | tuple[int, ...] | NDArray) -> Basis:
        """
        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define an additive basis
        >>> basis_1 = nmo.basis.BSplineEval(5)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(6)
        >>> basis_3 = nmo.basis.RaisedCosineLinearEval(7)
        >>> additive_basis = basis_1 + basis_2 + basis_3

        Specify the input shape using all 3 allowed ways: integer, tuple, array
        >>> _ = additive_basis.set_input_shape(1, (2, 3), np.ones((10, 4, 5)))

        Expected output features are:
        (5 bases * 1 input) + (6 bases * 6 inputs) + (7 bases * 20 inputs) = 181
        >>> additive_basis.n_output_features
        181

        """
        return super().set_input_shape(*xi)

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    @check_one_dimensional
    def _evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Evaluate the basis at the input samples.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs)

        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo
        >>> x, y = np.random.normal(size=(2, 30))

        >>> # define two basis objects and add them
        >>> basis_1 = nmo.basis.BSplineEval(10)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
        >>> additive_basis = basis_1 + basis_2

        >>> # call the basis.
        >>> out = additive_basis._evaluate(x, y)

        """
        X = np.hstack(
            (
                self.basis1._evaluate(*xi[: self.basis1._n_input_dimensionality]),
                self.basis2._evaluate(*xi[self.basis1._n_input_dimensionality :]),
            )
        )
        return X

    @add_docstring("compute_features", Basis)
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_add = basis1 + basis2
        >>> X_multi = basis_add.compute_features(np.random.randn(20), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 17 = 5 + 2*6
        (20, 17)

        """
        return super().compute_features(*xi)

    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Compute features for added bases and concatenate.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The features, shape (n_samples, n_basis_funcs)

        """
        # the numpy conversion is important, there is some in-place
        # array modification in basis.
        hstack_pynapple = support_pynapple(conv_type="numpy")(np.hstack)
        X = hstack_pynapple(
            (
                self.basis1._compute_features(
                    *xi[: self.basis1._n_input_dimensionality]
                ),
                self.basis2._compute_features(
                    *xi[self.basis1._n_input_dimensionality :]
                ),
            ),
        )
        return X

    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Decompose an array along a specified axis into sub-arrays based on the basis components.

        This function takes an array (e.g., a design matrix or model coefficients) and splits it along
        a designated axis. Each split corresponds to a different additive component of the basis,
        preserving all dimensions except the specified axis.

        **How It Works:**

        Suppose the basis is made up of **m components**, each with :math:`b_i` basis functions and :math:`n_i` inputs.
        The total number of features, :math:`N`, is calculated as:

        .. math::
            N = b_1 \cdot n_1 + b_2 \cdot n_2 + \ldots + b_m \cdot n_m

        This method splits any axis of length :math:`N` into sub-arrays, one for each basis component.

        The sub-array for the i-th basis component is reshaped into dimensions
        :math:`(n_i, b_i)`.

        For example, if the array shape is :math:`(1, 2, N, 4, 5)`, then each split sub-array will have shape:

        .. math::
            (1, 2, n_i, b_i, 4, 5)

        where:

        - :math:`n_i` represents the number of inputs associated with the i-th component,
        - :math:`b_i` represents the number of basis functions in that component.

        The specified axis (``axis``) determines where the split occurs, and all other dimensions
        remain unchanged. See the example section below for the most common use cases.

        Parameters
        ----------
        x :
            The input array to be split, representing concatenated features, coefficients,
            or other data. The shape of ``x`` along the specified axis must match the total
            number of features generated by the basis, i.e., ``self.n_output_features``.

            **Examples:**
            - For a design matrix: ``(n_samples, total_n_features)``
            - For model coefficients: ``(total_n_features,)`` or ``(total_n_features, n_neurons)``.

        axis : int, optional
            The axis along which to split the features. Defaults to 1.
            Use ``axis=1`` for design matrices (features along columns) and ``axis=0`` for
            coefficient arrays (features along rows). All other dimensions are preserved.

        Raises
        ------
        ValueError
            If the shape of ``x`` along the specified axis does not match ``self.n_output_features``.

        Returns
        -------
        dict
            A dictionary where:

            - **Keys**: Labels of the additive basis components.
            - **Values**: Sub-arrays corresponding to each component. Each sub-array has the shape:

            .. math::
                (..., n_i1, n_i2,..,n_im, b_i, ...)

            - ``(n_samples, n_i1, n_i2,..,n_im)``: is the shape of the input to the i-th basis.
            - ``b_i``: The number of basis functions for the i-th basis component.

            These sub-arrays are reshaped along the specified axis, with all other dimensions
            remaining the same.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineConv
        >>> from nemos.glm import GLM
        >>> # Define an additive basis
        >>> basis = (
        ...     BSplineConv(n_basis_funcs=5, window_size=10, label="feature_1") +
        ...     BSplineConv(n_basis_funcs=6, window_size=10, label="feature_2")
        ... )
        >>> # Generate a sample input array and compute features
        >>> x1, x2 = np.random.randn(20), np.random.randn(20)
        >>> X = basis.compute_features(x1, x2)
        >>> # Split the feature matrix along axis 1
        >>> split_features = basis.split_by_feature(X, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        feature_1: shape (20, 5)
        feature_2: shape (20, 6)
        >>> # If one of the basis components accepts multiple inputs, the resulting dictionary will be nested:
        >>> multi_input_basis = BSplineConv(n_basis_funcs=6, window_size=10,
        ... label="multi_input")
        >>> X_multi = multi_input_basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = multi_input_basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        multi_input, shape (20, 2, 6)
        >>> # the method can be used to decompose the glm coefficients in the various features
        >>> counts = np.random.poisson(size=20)
        >>> model = GLM().fit(X, counts)
        >>> split_coef = basis.split_by_feature(model.coef_, axis=0)
        >>> for feature, coef in split_coef.items():
        ...     print(f"{feature}: shape {coef.shape}")
        feature_1: shape (5,)
        feature_2: shape (6,)

        """
        return super().split_by_feature(x, axis=axis)

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        The i-th axis of the grid will be sampled with ``n_samples[i]`` equi-spaced points.
        The method uses numpy.meshgrid with ``indexing="ij"``, returning matrix indexing
        instead of the default cartesian indexing, see Notes.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of points in the uniformly spaced grid. The length of
            n_samples must equal the number of combined bases.

        Returns
        -------
        *Xs :
            A tuple of arrays containing the meshgrid values, one element for each of the n dimension of the grid,
            where n equals to the number of inputs.
            The size of ``Xs[i]`` is ``(n_samples[0], ... , n_samples[n])``.
        Y :
            The basis function evaluated at the samples,
            shape ``(n_samples[0], ... , n_samples[n], number of basis)``.

        Raises
        ------
        ValueError
            If the time point number is inconsistent between inputs or if the number of inputs doesn't match what
            the Basis object requires.
        ValueError
            If one of the n_samples is <= 0.

        Notes
        -----
        Setting ``indexing = 'ij'`` returns a meshgrid with matrix indexing. In the N-D case with inputs of size
        :math:`M_1,...,M_N`, outputs are of shape :math:`(M_1, M_2, M_3, ....,M_N)`.
        This differs from the numpy.meshgrid default, which uses Cartesian indexing.
        For the same input, Cartesian indexing would return an output of shape :math:`(M_2, M_1, M_3, ....,M_N)`.

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define two basis objects and add them
        >>> basis_1 = nmo.basis.BSplineEval(10)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
        >>> additive_basis = basis_1 + basis_2

        >>> # evaluate on a grid of 10 x 10 equi-spaced samples
        >>> X, Y, Z = additive_basis.evaluate_on_grid(10, 10)

        """
        return super().evaluate_on_grid(*n_samples)

    def _get_feature_slicing(
        self,
        n_inputs: Optional[tuple] = None,
        start_slice: Optional[int] = None,
        split_by_input: bool = True,
    ) -> Tuple[dict, int]:
        """
        Calculate and return the slicing for features based on the input structure.

        This method determines how to slice the features for different basis types.

        Parameters
        ----------
        n_inputs :
            The number of input basis for each component, by default it uses ``self._n_basis_input``.
        start_slice :
            The starting index for slicing, by default it starts from 0.
        split_by_input :
            Flag indicating whether to split the slicing by individual inputs or not.
            If ``False``, a single slice is generated for all inputs.

        Returns
        -------
        split_dict :
            Dictionary with keys as labels and values as slices representing
            the slicing for each input or additive component, if split_by_input equals to
            True or False respectively.
        start_slice :
            The updated starting index after slicing.

        See Also
        --------
        _get_default_slicing : Handles default slicing logic.
        _merge_slicing_dicts : Merges multiple slicing dictionaries, handling keys conflicts.
        """
        # Set default values for n_inputs and start_slice if not provided
        n_inputs = n_inputs or self._input_shape_product
        start_slice = start_slice or 0

        # If the instance is of AdditiveBasis type, handle slicing for the additive components

        split_dict, start_slice = self.basis1._get_feature_slicing(
            n_inputs[: len(self.basis1._input_shape_product)],
            start_slice,
            split_by_input=split_by_input,
        )
        sp2, start_slice = self.basis2._get_feature_slicing(
            n_inputs[len(self.basis1._input_shape_product) :],
            start_slice,
            split_by_input=split_by_input,
        )
        split_dict = self._merge_slicing_dicts(split_dict, sp2)
        return split_dict, start_slice

    def _merge_slicing_dicts(self, dict1: dict, dict2: dict) -> dict:
        """Merge two slicing dictionaries, handling key conflicts."""
        for key, val in dict2.items():
            if key in dict1:
                new_key = self._generate_unique_key(dict1, key)
                dict1[new_key] = val
            else:
                dict1[key] = val
        return dict1

    def __iter__(self):
        """Iterate over components."""
        for bas in self.basis1:
            yield bas
        for bas in self.basis2:
            yield bas

    def __len__(self):
        return len(self.basis1) + len(self.basis2)


class MultiplicativeBasis(CompositeBasisMixin, Basis):
    """
    Class representing the multiplication (external product) of two Basis objects.

    Parameters
    ----------
    basis1 :
        First basis object to multiply.
    basis2 :
        Second basis object to multiply.

    Examples
    --------
    >>> # Generate sample data
    >>> import numpy as np
    >>> import nemos as nmo
    >>> X = np.random.normal(size=(30, 3))

    >>> # define two basis and multiply
    >>> basis_1 = nmo.basis.BSplineEval(10)
    >>> basis_2 = nmo.basis.RaisedCosineLinearEval(15)
    >>> multiplicative_basis = basis_1 * basis_2
    >>> multiplicative_basis
    MultiplicativeBasis(
        basis1=BSplineEval(n_basis_funcs=10, order=4),
        basis2=RaisedCosineLinearEval(n_basis_funcs=15, width=2.0),
    )
    >>> # Can multiply or add another basis to the AdditiveBasis object
    >>> # This will cause the number of output features of the result basis to grow accordingly
    >>> basis_3 = nmo.basis.RaisedCosineLogEval(100)
    >>> multiplicative_basis_2 = multiplicative_basis * basis_3
    >>> multiplicative_basis_2
    MultiplicativeBasis(
        basis1=MultiplicativeBasis(
            basis1=BSplineEval(n_basis_funcs=10, order=4),
            basis2=RaisedCosineLinearEval(n_basis_funcs=15, width=2.0),
        ),
        basis2=RaisedCosineLogEval(n_basis_funcs=100, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True),
    )
    """

    def __init__(self, basis1: Basis, basis2: Basis) -> None:
        CompositeBasisMixin.__init__(self, basis1, basis2)
        Basis.__init__(self, mode="composite")
        self._label = "(" + basis1.label + " * " + basis2.label + ")"
        self._n_input_dimensionality = (
            basis1._n_input_dimensionality + basis2._n_input_dimensionality
        )

    @property
    def n_basis_funcs(self):
        """Compute the n-basis function runtime.

        This plays well with cross-validation where the number of basis function of the
        underlying bases can be changed. It must be read-only since the number of basis
        is determined by the two basis elements and the type of composition.
        """
        return self.basis1.n_basis_funcs * self.basis2.n_basis_funcs

    @property
    def n_output_features(self):
        out1 = getattr(self.basis1, "n_output_features", None)
        out2 = getattr(self.basis2, "n_output_features", None)
        if out1 is None or out2 is None:
            return None
        return out1 * out2

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    @check_one_dimensional
    def _evaluate(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor) -> FeatureMatrix:
        """
        Evaluate the basis at the input samples.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs)

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(5) * nmo.basis.RaisedCosineLinearEval(6)
        >>> x, y = np.random.randn(2, 30)
        >>> X = mult_basis._evaluate(x, y)
        """
        X = np.asarray(
            row_wise_kron(
                self.basis1._evaluate(*xi[: self.basis1._n_input_dimensionality]),
                self.basis2._evaluate(*xi[self.basis1._n_input_dimensionality :]),
                transpose=False,
            )
        )
        return X

    def _compute_features(
        self, *xi: NDArray | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Compute the features for the multiplied bases, and compute their outer product.

        Parameters
        ----------
        xi[0], ..., xi[n] : (n_samples,)
            Tuple of input samples, each with the same number of samples. The
            number of input arrays must equal the number of combined bases.

        Returns
        -------
        :
            The  features, shape (n_samples, n_basis_funcs)

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(5) * nmo.basis.RaisedCosineLinearEval(6)
        >>> x, y = np.random.randn(2, 30)
        >>> X = mult_basis.compute_features(x, y)
        """
        kron = support_pynapple(conv_type="numpy")(row_wise_kron)
        X = kron(
            self.basis1._compute_features(*xi[: self.basis1._n_input_dimensionality]),
            self.basis2._compute_features(*xi[self.basis1._n_input_dimensionality :]),
            transpose=False,
        )
        return X

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        The i-th axis of the grid will be sampled with ``n_samples[i]`` equi-spaced points.
        The method uses numpy.meshgrid with ``indexing="ij"``, returning matrix indexing
        instead of the default cartesian indexing, see Notes.

        Parameters
        ----------
        n_samples[0],...,n_samples[n]
            The number of points in the uniformly spaced grid. The length of
            n_samples must equal the number of combined bases.

        Returns
        -------
        *Xs :
            A tuple of arrays containing the meshgrid values, one element for each of the n dimension of the grid,
            where n equals to the number of inputs.
            The size of ``Xs[i]`` is ``(n_samples[0], ... , n_samples[n])``.
        Y :
            The basis function evaluated at the samples,
            shape ``(n_samples[0], ... , n_samples[n], number of basis)``.

        Raises
        ------
        ValueError
            If the time point number is inconsistent between inputs or if the number of inputs doesn't match what
            the Basis object requires.
        ValueError
            If one of the n_samples is <= 0.

        Notes
        -----
        Setting ``indexing = 'ij'`` returns a meshgrid with matrix indexing. In the N-D case with inputs of size
        :math:`M_1,...,M_N`, outputs are of shape :math:`(M_1, M_2, M_3, ....,M_N)`.
        This differs from the numpy.meshgrid default, which uses Cartesian indexing.
        For the same input, Cartesian indexing would return an output of shape :math:`(M_2, M_1, M_3, ....,M_N)`.

        Examples
        --------
        >>> import numpy as np
        >>> import nemos as nmo
        >>> mult_basis = nmo.basis.BSplineEval(4) * nmo.basis.RaisedCosineLinearEval(5)
        >>> X, Y, Z = mult_basis.evaluate_on_grid(10, 10)
        """
        return super().evaluate_on_grid(*n_samples)

    @add_docstring("compute_features", Basis)
    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_mul = basis1 * basis2
        >>> X_multi = basis_mul.compute_features(np.random.randn(20), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 60 = 5 * 2 * 6
        (20, 60)

        """
        return super().compute_features(*xi)

    @add_docstring("split_by_feature", Basis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval, RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis1 = BSplineEval(n_basis_funcs=5, label="one_input")
        >>> basis2 = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> basis_mul = basis1 * basis2
        >>> X_multi = basis_mul.compute_features(np.random.randn(20), np.random.randn(20, 2))
        >>> print(X_multi.shape) # num_features: 60 = 5 * 2 * 6
        (20, 60)

        >>> # The multiplicative basis is a single 2D component.
        >>> split_features = basis_mul.split_by_feature(X_multi, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        (one_input * two_inputs): shape (20, 2, 30)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", CompositeBasisMixin)
    def set_input_shape(self, *xi: int | tuple[int, ...] | NDArray) -> Basis:
        """
        Examples
        --------
        >>> # Generate sample data
        >>> import numpy as np
        >>> import nemos as nmo

        >>> # define an additive basis
        >>> basis_1 = nmo.basis.BSplineEval(5)
        >>> basis_2 = nmo.basis.RaisedCosineLinearEval(6)
        >>> basis_3 = nmo.basis.MSplineEval(7)
        >>> multiplicative_basis = basis_1 * basis_2 * basis_3

        Specify the input shape using all 3 allowed ways: integer, tuple, array
        >>> _ = multiplicative_basis.set_input_shape(1, (2, 3), np.ones((10, 4, 5)))

        Expected output features are:
        (5 * 6 * 7 bases) * (1 * 6 * 20 inputs) = 25200
        >>> multiplicative_basis.n_output_features
        25200

        """
        return super().set_input_shape(*xi)



import pytest
from myst_nb import glue

import nemos._documentation_utils._myst_nb_glue as myst_utils


@pytest.fixture
def functions():

    def func(x, y, z):
        # func body
        return x + y + z

    def func_multiline_init(
        long_variable_names,
        multi_line_init_with_black,
        multi_line_init_with_black_1,
        multi_line_init_with_black_3,
    ):
        # func_multiline_init body
        return

    def func_multiline_return(
        long_variable_names,
        multi_line_init_with_black,
        multi_line_init_with_black_1,
        multi_line_init_with_black_3,
    ):
        # func_multiline_return body
        return (
            long_variable_names,
            multi_line_init_with_black,
            multi_line_init_with_black_1,
            multi_line_init_with_black_3,
        )

    @myst_utils.capture_print
    def func_decorated(
        long_variable_names,
        multi_line_init_with_black,
        multi_line_init_with_black_1,
        multi_line_init_with_black_3,
    ):
        # func_decorated body
        return (
            long_variable_names,
            multi_line_init_with_black,
            multi_line_init_with_black_1,
            multi_line_init_with_black_3,
        )

    return (
        (func, "# func body"),
        (func_multiline_init, "# func_multiline_init body"),
        (func_multiline_return, "# func_multiline_return body"),
        (func_decorated, "# func_decorated body"),
    )


def test_parse_func_body(functions):
    for func, expected in functions:
        assert myst_utils.extract_body_exclude_def_and_return(func) == expected


def test_capture_out():
    @myst_utils.capture_print
    def func(x):
        print(x)
        return 2 * x

    out, doubled = func(4)
    assert doubled == 8
    assert out == "4\n"


def test_glue_formatted_str():
    out = myst_utils.FormattedString("string")
    capture_glue = myst_utils.capture_print(glue)
    out, _ = capture_glue("out-string", out)
    assert out == "string\n"
    out2, _ = capture_glue("not-formatted", "string")
    assert out2 == "'string'\n"


def test_gluing_does_not_display():
    func = myst_utils.capture_print(myst_utils.glue_two_step_convolve)
    out, _ = func()
    assert out == ""
